"""
This module provides the ClusterExpansion class.
"""

import pandas as pd
import numpy as np
import pickle
import tempfile
import tarfile
import re

from icet import ClusterSpace
from icet.core.structure import Structure
from typing import List, Union
from ase import Atoms


class ClusterExpansion:
    """Cluster expansions are obtained by combining a cluster space with
    a set of effective cluster interactions (ECIs). Instances of this
    class allow one to predict the property of interest for a given
    structure.

    Attributes
    ----------
    cluster_space : icet.ClusterSpace
        cluster space that was used for constructing the cluster expansion
    parameters : np.ndarray
        effective cluster interactions (ECIs)

    Example
    -------
    The following snippet illustrates the initialization and usage of
    a ClusterExpansion object. Here, the ECIs are taken to be a list
    of ones. Usually, they would be obtained by training with
    respect to a set of reference data::

        from ase.build import bulk
        from icet import ClusterSpace, ClusterExpansion

        # create cluster expansion
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[7.0, 5.0],
                          chemical_symbols=[['Au', 'Pd']])
        ecis = 14 * [1.0]
        ce = ClusterExpansion(cs, ecis)

        # make prediction for supercell
        sc = prim.repeat(3)
        for k in [1, 4, 7]:
            sc[k].symbol = 'Pd'
        print(ce.predict(sc))
    """

    def __init__(self, cluster_space: ClusterSpace, parameters: np.array,
                 metadata: dict = None) -> None:
        """
        Initializes a ClusterExpansion object.

        Parameters
        ----------
        cluster_space
            cluster space to be used for constructing the cluster expansion
        parameters
            effective cluster interactions (ECIs)
        metadata : dict
            metadata dictionary, user-defined metadata to be stored together
            with cluster expansion. Will be pickled when CE is written to file.
            By default contains icet version, username, hostname and date.

        Raises
        ------
        ValueError
            if cluster space and parameters differ in length
        """
        if len(cluster_space) != len(parameters):
            raise ValueError('cluster_space ({}) and parameters ({}) must have'
                             ' the same length'.format(len(cluster_space), len(parameters)))
        self._cluster_space = cluster_space.copy()
        if isinstance(parameters, list):
            parameters = np.array(parameters)
        self._parameters = parameters
        self._original_parameters = parameters.copy()

        # add metadata
        if metadata is None:
            metadata = dict()
        self._metadata = metadata
        self._add_default_metadata()

    def predict(self, structure: Union[Atoms, Structure]) -> float:
        """
        Predicts the property of interest (e.g., the energy) for the input
        structure using the cluster expansion.

        Parameters
        ----------
        structure
            atomic configuration

        Returns
        -------
        float
            property value of predicted by the cluster expansion
        """
        cluster_vector = self._cluster_space.get_cluster_vector(structure)
        prop = np.dot(cluster_vector, self.parameters)
        return prop

    def get_cluster_space_copy(self) -> ClusterSpace:
        """ Gets copy of cluster space on which cluster expansion is based """
        return self._cluster_space.copy()

    @property
    def parameters_as_dataframe(self) -> pd.DataFrame:
        """ dataframe containing orbit data and ECIs """
        rows = self._cluster_space.orbit_data
        for row, eci in zip(rows, self.parameters):
            row['eci'] = eci
        return pd.DataFrame(rows)

    @property
    def orders(self) -> List[int]:
        """ orders included in cluster expansion """
        return list(range(len(self._cluster_space.cutoffs) + 2))

    @property
    def parameters(self) -> List[float]:
        """ effective cluster interactions (ECIs) """
        return self._parameters

    @property
    def metadata(self):
        """ dict : metadata associated with cluster expansion """
        return self._metadata

    def plot_parameters(self, orders=None):
        """ Plot ECIs for given orders, default plots for all orders """

        if orders is None:
            orders = self.orders
        df = self.parameters_as_dataframe

        # plotting
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.axhline(y=0.0, c='k', lw=1)
        for order in orders:
            df_order = df.loc[df['order'] == order]
            ax.plot(df_order.radius, df_order.eci, 'o', ms=8, label='order {}'.format(order))
        ax.legend(loc='best')
        ax.set_xlabel('Radius')
        ax.set_ylabel('ECI')
        plt.show()

    def __len__(self) -> int:
        return len(self._parameters)

    def _get_string_representation(self, print_threshold: int = None,
                                   print_minimum: int = 10):
        """ String representation of the cluster expansion. """
        cluster_space_repr = self._cluster_space._get_string_representation(
            print_threshold, print_minimum).split('\n')
        # rescale width
        eci_col_width = max(len('{:9.3g}'.format(max(self._parameters, key=abs))), len('ECI'))
        width = len(cluster_space_repr[0]) + len(' | ') + eci_col_width

        s = []  # type: List
        s += ['{s:=^{n}}'.format(s=' Cluster Expansion ', n=width)]
        s += [t for t in cluster_space_repr if re.search(':', t)]

        # additional information about number of nonzero the ECIs
        df = self.parameters_as_dataframe
        orders = self.orders
        nzp_by_order = [np.count_nonzero(df[df.order == order].eci) for order in orders]
        assert sum(nzp_by_order) == np.count_nonzero(self.parameters)
        s += [' total number of nonzero parameters: {}'.format(sum(nzp_by_order))]
        line = ' number of nonzero parameters by order: '
        for order, nzp in zip(orders, nzp_by_order):
            line += '{}= {}  '.format(order, nzp)
        s += [line]

        # table header
        s += [''.center(width, '-')]
        t = [t for t in cluster_space_repr if 'index' in t]
        t += ['{s:^{n}}'.format(s='ECI', n=eci_col_width)]
        s += [' | '.join(t)]
        s += [''.center(width, '-')]

        # table body
        index = 0
        while index < len(self):
            if (print_threshold is not None and
                    len(self) > print_threshold and
                    index >= print_minimum and
                    index <= len(self) - print_minimum):
                index = len(self) - print_minimum
                s += [' ...']
            pattern = r'^{:4}'.format(index)
            t = [t for t in cluster_space_repr if re.match(pattern, t)]
            eci_value = '{:9.3g}'.format(self._parameters[index])
            t += ['{s:^{n}}'.format(s=eci_value, n=eci_col_width)]
            s += [' | '.join(t)]
            index += 1
        s += [''.center(width, '=')]

        return '\n'.join(s)

    def __repr__(self) -> str:
        """ string representation """
        return self._get_string_representation(print_threshold=50)

    def print_overview(self,
                       print_threshold: int = None,
                       print_minimum: int = 10) -> None:
        """
        Print an overview of the cluster expansion in terms of the orbits (order,
        radius, multiplicity, corresponding ECI etc).

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the orbit
            list if `print_threshold` is exceeded
        """
        print(self._get_string_representation(print_threshold=print_threshold,
                                              print_minimum=print_minimum))

    def prune(self, indices: List[int] = None, tol: float = 0):
        """
        Removes orbits from the cluster expansion (CE), for which the effective
        cluster interactions (ECIs; parameters) are zero or close to zero.
        This commonly reduces the computational cost for evaluating the CE and
        is therefore recommended prior to using it in production. If the method
        is called without arguments orbits will be pruned, for which the ECIs
        are strictly zero. Less restrictive pruning can be achived by setting
        the `tol` keyword.

        Parameters
        ----------
        indices
            indices to parameters to remove in the cluster expansion.
        tol
            orbits for which the absolute ECIs is/are within this
            value will be pruned
        """

        # find orbit indices to be removed
        if indices is None:
            indices = [i for i, param in enumerate(
                self.parameters) if np.abs(param) <= tol and i > 0]
        df = self.parameters_as_dataframe
        indices = list(set(indices))

        if 0 in indices:
            raise ValueError('Orbit index cannot be 0 since the zerolet may not be pruned.')
        orbit_candidates_for_removal = df.orbit_index[np.array(indices)].tolist()
        safe_to_remove_orbits, safe_to_remove_params = [], []
        for oi in set(orbit_candidates_for_removal):
            if oi == -1:
                continue
            orbit_count = df.orbit_index.tolist().count(oi)
            oi_remove_count = orbit_candidates_for_removal.count(oi)
            if orbit_count <= oi_remove_count:
                safe_to_remove_orbits.append(oi)
                safe_to_remove_params += df.index[df['orbit_index'] == oi].tolist()

        # prune cluster space
        self._cluster_space._prune_orbit_list(indices=safe_to_remove_orbits)
        self._parameters = self._parameters[np.setdiff1d(
            np.arange(len(self._parameters)), safe_to_remove_params)]
        assert len(self._parameters) == len(self._cluster_space)

    def write(self, filename: str):
        """
        Writes ClusterExpansion object to file.

        Parameters
        ---------
        filename
            name of file to which to write
        """
        self._cluster_space.write(filename)

        items = dict()
        items['parameters'] = self.parameters

        # TODO: remove if condition once metadata is firmly established
        if hasattr(self, '_metadata'):
            items['metadata'] = self._metadata

        with tarfile.open(name=filename, mode='w') as tar_file:
            cs_file = tempfile.NamedTemporaryFile()
            self._cluster_space.write(cs_file.name)
            tar_file.add(cs_file.name, arcname='cluster_space')

            # write items
            temp_file = tempfile.TemporaryFile()
            pickle.dump(items, temp_file)
            temp_file.seek(0)
            tar_info = tar_file.gettarinfo(arcname='items', fileobj=temp_file)
            tar_file.addfile(tar_info, temp_file)
            temp_file.close()

    @staticmethod
    def read(filename: str):
        """
        Reads ClusterExpansion object from file.

        Parameters
        ---------
        filename
            file from which to read
        """
        with tarfile.open(name=filename, mode='r') as tar_file:
            cs_file = tempfile.NamedTemporaryFile()
            cs_file.write(tar_file.extractfile('cluster_space').read())
            cs_file.seek(0)
            cs = ClusterSpace.read(cs_file.name)
            items = pickle.load(tar_file.extractfile('items'))

        parameters = items['parameters']
        ce = ClusterExpansion(cs, parameters)

        # TODO: remove if condition once metadata is firmly established
        if 'metadata' in items:
            ce._metadata = items['metadata']
        else:
            del ce._metadata

        assert list(parameters) == list(ce.parameters)
        return ce

    def _add_default_metadata(self):
        """Adds default metadata to metadata dict."""
        import getpass
        import socket
        from datetime import datetime
        from icet import __version__ as icet_version

        self._metadata['date_created'] = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self._metadata['username'] = getpass.getuser()
        self._metadata['hostname'] = socket.gethostname()
        self._metadata['icet_version'] = icet_version



"""
This module provides a Python interface to the LatticeSite class.
"""

from _icet import LatticeSite


def __latticeSite_repr(self):
    return '{} : {}'.format(self.index, self.unitcell_offset)


LatticeSite.__repr__ = __latticeSite_repr



"""
This module provides a Python interface to the NeighborList class
with supplementary functions.
"""

from typing import List, Union
from _icet import NeighborList
from ase import Atoms
from .structure import Structure


def get_neighbor_lists(structure: Union[Atoms, Structure],
                       cutoffs: List[float] = None) -> List[NeighborList]:
    """
    Returns a list of icet neighbor lists given a configuration and cutoffs.

    Parameters
    ----------
    structure
        atomic configuration
    cutoffs
        positive floats indicating the cutoffs for the various clusters
    """

    # deal with different types of structure objects
    if isinstance(structure, Atoms):
        structure = Structure.from_atoms(structure)
    elif not isinstance(structure, Structure):
        msg = ['Unknown structure format']
        msg += ['{} (ClusterSpace)'.format(type(structure))]
        raise Exception(' '.join(msg))

    neighbor_lists = []
    if cutoffs is None:
        raise Exception('Both n and cutoffs is None in count clusters')
    else:
        for cutoff in cutoffs:
            nl = NeighborList(cutoff)
            neighbor_lists.append(nl)

    # build the neighbor_lists
    for nl in neighbor_lists:
        nl.build(structure)

    return neighbor_lists



import logging
import sys

# This is the root logger of icet
logger = logging.getLogger('icet')

# Set the format which is simpler for the console
FORMAT = '%(name)s: %(levelname)s  %(message)s'
formatter = logging.Formatter(FORMAT)

# Will process all levels of DEBUG or higher
logger.setLevel(logging.DEBUG)

# If you know what you are doing you may set this to True
logger.propagate = False

# The logger will collect events from children and the default
# behaviour is to print it directly to stdout
sh = logging.StreamHandler(sys.stdout)
sh.setFormatter(formatter)
sh.setLevel(logging.WARNING)
logger.addHandler(sh)


class MyFilter(logging.Filter):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno == self.__level


def set_log_config(filename=None, level=None, restricted=False):

    # If a filename is provided a logfile will be created
    if filename is not None:
        fh = logging.FileHandler(filename)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

        # Restrict file handler to only write out at the level specified
        if restricted:
            lvl = logging.getLevelName(level)
            fh.addFilter(MyFilter(lvl))

        if level is not None:
            fh.setLevel(level)

    # If only level is provided, the stream handler will be reset to this level
    elif level is not None:
        sh.setLevel(level)



"""
Definition of the variance-constrained semi-grand canonical ensemble class.
"""

import numpy as np

from ase import Atoms
from ase.data import atomic_numbers, chemical_symbols
from ase.units import kB
from typing import Dict, Union, List

from .. import DataContainer
from ..calculators.base_calculator import BaseCalculator
from .thermodynamic_base_ensemble import ThermodynamicBaseEnsemble


class VCSGCEnsemble(ThermodynamicBaseEnsemble):
    """Instances of this class allow one to simulate systems in the
    variance-constrained semi-grand canonical (VCSGC) ensemble
    (:math:`N\\phi\\kappa VT`), i.e. at constant temperature (:math:`T`), total
    number of sites (:math:`N=\\sum_i N_i`), and two additional dimensionless
    parameters :math:`\\phi` and :math:`\\kappa`, which constrain average and
    variance of the concentration, respectively.

    The below examples treat the binary case, but the generalization of
    to ternaries and higher-order systems is straight-forward.
    The probability for a particular state in the VCSGC ensemble for a
    :math:`2`-component system can be written

    .. math::

        \\rho_{\\text{VCSGC}} \\propto \\exp\\Big[ - E / k_B T
        + \\kappa N ( c_1 + \\phi_1 / 2 )^2 \\Big],

    where :math:`c_1` represents the concentration of species 1, i.e.
    :math:`c_1=N_1/N`. (Please note that the quantities :math:`\\kappa` and
    :math:`\\phi` correspond, respectively, to :math:`\\bar{\\kappa}` and
    :math:`\\bar{\\phi}` in [SadErh12]_.). The :math:`\\phi` may refer to any
    of the two species. If :math:`\\phi` is specified for species A, an
    equivalent simulation can be carried out by specifying :math:`\\phi_B` as
    :math:`-2-\\phi_A`. In general, simulations of :math:`N`-component
    systems requires the specification of :math:`\\phi` for :math:`N-1`
    elements.

    Just like the :ref:`semi-grand canonical ensemble <canonical_ensemble>`,
    the VCSGC ensemble allows concentrations to change. A trial step consists
    of changing the identity of a randomly chosen atom and accepting the change
    with probability

    .. math::

        P = \\min \\{ 1, \\, \\exp [ - \\Delta E / k_B T
        + \\kappa N \\Delta c_1 (\\phi_1 + \\Delta c_1 + 2 c_1 ) ] \\}.

    Note that for a sufficiently large value of :math:`\\kappa`, say 200, the
    probability density :math:`\\rho_{\\text{VCSGC}}` is sharply peaked around
    :math:`c_1=-\\phi_1 / 2`. In practice, this means that we can gradually
    change :math:`\\phi_1` from (using some margins) :math:`-2.1` to
    :math:`0.1` and take the system continuously from :math:`c_1 = 0` to
    :math:`1`. The parameter :math:`\\kappa` constrains the fluctuations (or
    the variance) of the concentration at each value of :math:`\\phi_1`, with
    higher values of :math:`\\kappa` meaning less fluctuations. Unlike the
    :ref:`semi-grand canonical ensemble <vcsgc_ensemble>`, one value of
    :math:`\\phi_1` maps to one and only one concentration also in multiphase
    regions. Since the derivative of the canonical free energy can be expressed
    in terms of parameters and observables of the VCSGC ensemble,

    .. math::

        k_B T \\kappa ( \\phi_1 + 2 \\langle c_1 \\rangle ) = - \\frac{1}{N}
        \\frac{\\partial F}{\\partial c_1} (N, V, T, \\langle c_1 \\rangle ),

    this ensemble allows for thermodynamic integration across multiphase
    regions. This means that we can construct phase diagrams by directly
    comparing the free energies of the different phases. This often makes the
    VCSGC ensemble more convenient than the :ref:`semi-grand canonical ensemble
    <sgc_ensemble>` when simulating materials with multiphase regions, such as
    alloys with miscibility gaps.

    When using the VCSGC ensemble, please cite Sadigh, B. and Erhart, P., Phys.
    Rev. B **86**, 134204 (2012) [SadErh12]_.

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    temperature : float
        temperature :math:`T` in appropriate units [commonly Kelvin]
    phis : Dict[str, float]
        average constraint parameters :math:`\\phi_i`; the key denotes the
        species; for a N-component sublattice, there should be N - 1
        different `\\phi_i` (referred to as :math:`\\bar{\\phi}`
        in [SadErh12]_)
    kappa : float
        parameter that constrains the variance of the concentration
        (referred to as :math:`\\bar{\\kappa}` in [SadErh12]_)
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    sublattice_probabilities : List[float]
        probability for picking a sublattice when doing a random flip.
        The list should be as long as the number of sublattices and should
        sum up to 1.


    Example
    -------
    The following snippet illustrate how to carry out a simple Monte Carlo
    simulation in the variance-constrained semi-canonical ensemble. Here, the
    parameters of the cluster expansion are set to emulate a simple Ising model
    in order to obtain an example that can be run without modification. In
    practice, one should of course use a proper cluster expansion::

        from ase.build import bulk
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import VCSGCEnsemble

        # prepare cluster expansion
        # the setup emulates a second nearest-neighbor (NN) Ising model
        # (zerolet and singlet ECIs are zero; only first and second neighbor
        # pairs are included)
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[4.3], chemical_symbols=['Ag', 'Au'])
        ce = ClusterExpansion(cs, [0, 0, 0.1, -0.02])

        # set up and run MC simulation
        structure = prim.repeat(3)
        calc = ClusterExpansionCalculator(structure, ce)
        phi = 0.6
        mc = VCSGCEnsemble(structure=structure, calculator=calc,
                           temperature=600,
                           data_container='myrun_vcsgc.dc',
                           phis={'Au': phi},
                           kappa=200)
        mc.run(100)  # carry out 100 trial swaps
    """

    def __init__(self, structure: Atoms,
                 calculator: BaseCalculator,
                 temperature: float,
                 phis: Dict[str, float],
                 kappa: float,
                 boltzmann_constant: float = kB,
                 user_tag: str = None,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None,
                 sublattice_probabilities: List[float] = None) -> None:

        self._ensemble_parameters = dict(temperature=temperature,
                                         kappa=kappa)
        self._boltzmann_constant = boltzmann_constant

        # Save ensemble parameters
        for sym, phi in phis.items():
            if isinstance(sym, str):
                chemical_symbol = sym
            else:
                chemical_symbol = chemical_symbols[sym]
            phi_sym = 'phi_{}'.format(chemical_symbol)
            self._ensemble_parameters[phi_sym] = phi

        super().__init__(
            structure=structure,
            calculator=calculator,
            user_tag=user_tag,
            random_seed=random_seed,
            data_container=data_container,
            data_container_class=DataContainer,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval,
            boltzmann_constant=boltzmann_constant
        )

        # Save phis (need self.configuration to check sublattices so
        # we do it last)
        self._phis = get_phis(phis)

        # Check that each sublattice has N - 1 phis
        for sl in self.sublattices.active_sublattices:
            count_specified_elements = 0
            for number in sl.atomic_numbers:
                if number in self._phis.keys():
                    count_specified_elements += 1
            if count_specified_elements != len(sl.atomic_numbers) - 1:
                raise ValueError('phis must be set for N - 1 elements on a '
                                 'sublattice with N species')

        if sublattice_probabilities is None:
            self._flip_sublattice_probabilities = self._get_flip_sublattice_probabilities()
        else:
            self._flip_sublattice_probabilities = sublattice_probabilities

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """
        sublattice_index = self.get_random_sublattice_index(
            probability_distribution=self._flip_sublattice_probabilities)
        return self.do_vcsgc_flip(
            phis=self.phis, kappa=self.kappa, sublattice_index=sublattice_index)

    @property
    def temperature(self) -> float:
        """ temperature :math:`T` (see parameters section above) """
        return self.ensemble_parameters['temperature']

    @property
    def phis(self) -> Dict[int, float]:
        """
        phis :math:`\\phi_i`; one for each species but their sum must be
        :math:`-2.0` (referred to as :math:`\\bar{\\phi}` in [SadErh12]_)
        """
        return self._phis

    @property
    def kappa(self) -> float:
        """
        kappa :math:`\\bar{\\kappa}` constrain parameter
        (see parameters section above)
        """
        return self.ensemble_parameters['kappa']

    def _get_ensemble_data(self) -> Dict:
        """
        Returns a dict with the default data of the ensemble. This includes
        atom counts and free energy derivative.
        """
        data = super()._get_ensemble_data()

        # free energy derivative
        data.update(self._get_vcsgc_free_energy_derivatives(self.phis, self.kappa))

        # species counts
        data.update(self._get_species_counts())

        return data


def get_phis(phis: Dict[Union[int, str], float]) -> Dict[int, float]:
    """Get phis as used in the vcsgc ensemble.

    Parameters
    ----------
    phis
        the phis that will be transformed to the format
        used by the ensemble
    """
    if not isinstance(phis, dict):
        raise TypeError('phis has the wrong type: {}'.format(type(phis)))

    # Translate to atomic numbers if necessary
    phis_ret = {}
    for key, phi in phis.items():
        if isinstance(key, str):
            atomic_number = atomic_numbers[key]
            phis_ret[atomic_number] = phi
        elif isinstance(key, int):
            phis_ret[key] = phi

    return phis_ret



from .cluster_expansion_observer import ClusterExpansionObserver
from .cluster_count_observer import ClusterCountObserver
from .binary_short_range_order_observer import BinaryShortRangeOrderObserver
from .site_occupancy_observer import SiteOccupancyObserver

__all__ = ['ClusterExpansionObserver',
           'ClusterCountObserver',
           'BinaryShortRangeOrderObserver',
           'SiteOccupancyObserver']



from abc import ABC, abstractmethod
from ase import Atoms
from typing import List


class BaseCalculator(ABC):
    """
    Base class for calculators.

    Attributes
    ----------
    name : str
        human-readable calculator name
    """

    def __init__(self, structure, name='BaseCalculator'):
        self._structure = structure.copy()
        self._structure.wrap()
        self.name = name

    @property
    def structure(self) -> Atoms:
        """ atomic structure associated with calculator """
        return self._structure

    @abstractmethod
    def calculate_total(self):
        pass

    @abstractmethod
    def calculate_local_contribution(self):
        pass

    def update_occupations(self, indices: List[int], species: List[int]):
        """Updates the occupation (species) of the associated atomic
        structure.

        Parameters
        ----------
        indices
            sites to update
        species
            new occupations (species) by atomic number
        """
        if not isinstance(indices, list) and not isinstance(species, list):
            raise TypeError('sites and species must be of type list')
        if len(indices) != len(species):
            raise ValueError('sites and species must have the same length')
        self._structure.numbers[indices] = species

    @property
    def sublattices(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-

from .configuration_manager import ConfigurationManager
from .data_containers.data_container import DataContainer
from .data_containers.wang_landau_data_container import WangLandauDataContainer

"""
mchammer - Monte Carlo simulation module
"""

__project__ = 'icet-mchammer'
__description__ = 'icet Monte Carlo simulations module'
__all__ = ['ConfigurationManager',
           'DataContainer',
           'WangLandauDataContainer']
__maintainer__ = 'The icet developers team'
__maintainer_email__ = 'icet@materialsmodeling.org'
__url__ = 'http://icet.materialsmodeling.org/'



""" Data container class. """

import numpy as np
from .base_data_container import BaseDataContainer
from ..data_analysis import analyze_data


class DataContainer(BaseDataContainer):
    """
    Data container for storing information concerned with
    Monte Carlo simulations performed with mchammer.

    Parameters
    ----------
    structure : ase.Atoms
        reference atomic structure associated with the data container

    ensemble_parameters : dict
        parameters associated with the underlying ensemble

    metadata : dict
        metadata associated with the data container
    """

    def analyze_data(self, tag: str, start: int = None,
                     stop: int = None, max_lag: int = None) -> dict:
        """
        Returns detailed analysis of a scalar observerable.

        Parameters
        ----------
        tag
            tag of field over which to average
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.
        max_lag
            maximum lag between two points in data series, by default the
            largest length of the data series will be used.
            Used for computing autocorrelation
        Raises
        ------
        ValueError
            if observable is requested that is not in data container
        ValueError
            if observable is not scalar
        ValueError
            if observations is not evenly spaced

        Returns
        -------
        dict
            calculated properties of the data including mean,
            standard_deviation, correlation_length and error_estimate
            (95% confidence)
        """
        if tag in ['trajectory', 'occupations']:
            raise ValueError('{} is not scalar'.format(tag))
        steps, data = self.get_data('mctrial', tag, start=start, stop=stop)

        # check that steps are evenly spaced
        diff = np.diff(steps)
        step_length = diff[0]
        if not np.allclose(step_length, diff):
            raise ValueError('data records must be evenly spaced.')

        summary = analyze_data(data, max_lag=max_lag)
        summary['correlation_length'] *= step_length  # in mc-trials
        return summary

    def get_average(self, tag: str, start: int = None, stop: int = None) -> float:
        """
        Returns average of a scalar observable.

        Parameters
        ----------
        tag
            tag of field over which to average
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.

        Raises
        ------
        ValueError
            if observable is requested that is not in data container
        ValueError
            if observable is not scalar
        """
        if tag in ['trajectory', 'occupations']:
            raise ValueError('{} is not scalar'.format(tag))
        data = self.get_data(tag, start=start, stop=stop)
        return np.mean(data)



from itertools import product

import numpy as np

from ase import Atoms
from ase.build import make_supercell
from spglib import get_symmetry
from spglib import niggli_reduce as spg_nigg_red
from .structure_enumeration_support.hermite_normal_form \
    import yield_reduced_hnfs, HermiteNormalForm
from .structure_enumeration_support.smith_normal_form \
    import get_unique_snfs, SmithNormalForm
from .structure_enumeration_support.labeling_generation \
    import LabelingGenerator
from typing import Dict, List, Tuple, Generator


def _translate_labelings(
        labeling: tuple,
        snf: SmithNormalForm,
        nsites: int,
        include_self: bool = False) -> Generator[Tuple[int], None, None]:
    """
    Yields labelings that are equivalent to the original labeling under
    translations as dictated by the Smith normal form (SNF) provided.

    Parameters
    ----------
    labeling
        labeling to be translated
    snf
        SmithNormalForm object
    nsites
        number of sites in the primitive cell
    include_self
        if True original labeling will be included
    """

    # Compute size of each block within which translations occur
    sizes = [nsites * block for block in snf.blocks]

    # Loop over all possible translations within group as defined by snf
    for trans in product(range(snf.S[0]), range(snf.S[1]), range(snf.S[2])):
        if not include_self and sum(trans) == 0:
            continue

        labeling_trans = ()
        for i in range(snf.S[0]):
            group = (i + trans[0]) % snf.S[0]
            block_i = labeling[sizes[0] * group:sizes[0] * (group + 1)]
            for j in range(snf.S[1]):
                group = (j + trans[1]) % snf.S[1]
                block_j = block_i[sizes[1] * group:sizes[1] * (group + 1)]
                for k in range(snf.S[2]):
                    group = (k + trans[2]) % snf.S[2]
                    labeling_trans += tuple(block_j[sizes[2] * group:
                                                    sizes[2] * (group + 1)])
        yield labeling_trans


def _get_all_labelings(snf: SmithNormalForm,
                       labeling_generator: LabelingGenerator,
                       nsites: int) -> List[tuple]:
    """
    Returns inequivalent labelings corresponding to a Smith normal form
    (SNF) matrix. Superperiodic labelings as well as labelings that are
    equivalent under translations for this particular SNF will not be
    included. However, labelings that are equivalent by rotations that
    leave the cell (but not the labeling) unchanged will still be
    included, since these have to be removed for each Hermite normal
    form (HNF) separately.

    Parameters
    ----------
    snf
        SmithNormalForm object
    labeling_generator
        LabelingGenerator object
    nsites
        number of sites per primitive cell
    """
    labelings = []
    for labeling in labeling_generator.yield_labelings(snf.ncells):
        unique = True
        for labeling_trans in _translate_labelings(labeling, snf, nsites,
                                                   include_self=False):
            # Check whether it translates into itself. If so,
            # then it has been added with a smaller cell.
            if labeling == labeling_trans:
                unique = False
                break

            # Check with previous labelings,
            # if labeling can be translated into a previously
            # added labeling, then it is not unique
            if labeling_trans in labelings:
                unique = False
                break
        if unique:
            labelings.append(labeling)
    return labelings


def _permute_labeling(labeling: tuple, snf: SmithNormalForm,
                      transformation: List[np.ndarray],
                      nsites: int) -> Tuple[int]:
    """
    Returns permuted labeling according to transformations defined by
    transformation.

    Parameters
    ----------
    labeling
        labeling to be rotated
    snf
        SmithNormalForm object
    transformation
        transformation consisting of rotation, translation and basis
        shift
    nsites
        number of sites in the primitive cell
    """

    # Calculate transformation imposed by LRL multiplication
    new_group_order = np.dot(snf.group_order, transformation[0].T)

    # Loop over every atom to find its new position
    labeling_new = [0] * len(labeling)
    for member_index, member in enumerate(new_group_order):

        # Transform according to Gp,
        # but each site in the primitive cell also transforms in its own way
        for basis in range(nsites):
            new_cell = member + transformation[1][basis]

            # Calculate new index, first by finding the right block,
            # then by adding the basis index to that block
            new_index = 0
            for i in range(3):
                new_index += (new_cell[i] % snf.S[i]) * snf.blocks[i] * nsites
            new_index += transformation[2][basis]

            # Add the contribution to the hash key
            element = labeling[member_index * nsites + basis]
            labeling_new[new_index] = element
    return tuple(labeling_new)


def _yield_unique_labelings(labelings: List[int], snf: SmithNormalForm,
                            hnf: HermiteNormalForm, nsites: int) -> tuple:
    """
    Yields labelings that are unique in every imaginable sense.

    Parameters
    ----------
    labelkeys
        list of hash keys to labelings that may still contain labelings
        that are equivalent under rotations that leave the supercell
        shape unchanged
    snf
        SmithNormalForm object
    hnf
        HermiteNormalForm object
    nsites
        number of sites in the primitive cell
    """
    saved_labelings = []
    for labeling in labelings:

        # Check whether labeling is just a rotated version of a previous
        # labeling. Apply transformation that is specific to the hnf
        # and check all translations of the transformed labeling.
        unique = True
        for transformation in hnf.transformations:

            labeling_rot = _permute_labeling(labeling, snf, transformation,
                                             nsites)

            # Commonly, the transformation leaves the labeling
            # unchanged, so check that first as a special case
            # (yields a quite significant speedup)
            if labeling_rot == labeling:
                continue

            # Translate in all possible ways
            for labeling_rot_trans in \
                    _translate_labelings(labeling_rot, snf, nsites,
                                         include_self=True):
                if labeling_rot_trans in saved_labelings:
                    # Then we have rotated and translated the labeling
                    # into one that was already yielded
                    unique = False
                    break
            if not unique:
                break
        if unique:
            # Then we have finally found a unique structure
            # defined by an HNF matrix and a labeling
            saved_labelings.append(labeling)
            yield labeling


def _labeling_to_structure(labeling: tuple, hnf: np.ndarray, cell: np.ndarray,
                           new_cell: np.ndarray,
                           basis: np.ndarray, chemical_symbols: List[str],
                           pbc: List[bool]) -> Atoms:
    """
    Returns structure object corresponding to the given labeling using
    labeling, HNF matrix and parent lattice.

    Parameters
    ---------
    labeling
        permutation of index of elements
    hnf
        HNF object defining the supercell
    cell
        basis vectors of primtive cell listed row-wise
    new_cell
        new cell shape
    basis
        scaled coordinates to all sites in the primitive cell
    chemical_symbols
        list of elements, e.g. ``['Au', 'Ag']``
    pbc
        periodic boundary conditions of the primitive structure
    """
    symbols = []
    positions = []
    count = 0
    for i in range(hnf.H[0, 0]):
        coord = i * hnf.H[1, 0]
        offset10 = coord // hnf.H[0, 0] + coord % hnf.H[0, 0]
        coord = i * hnf.H[2, 0]
        offset20 = coord // hnf.H[0, 0] + coord % hnf.H[0, 0]
        for j in range(hnf.H[1, 1]):
            coord = j * hnf.H[2, 1]
            offset21 = coord // hnf.H[1, 1] + coord % hnf.H[1, 1]
            for k in range(hnf.H[2, 2]):
                for basis_vector in basis:
                    positions.append(i * cell[0] +
                                     (j + offset10) * cell[1] +
                                     (k + offset20 + offset21) * cell[2] +
                                     np.dot(cell.T, basis_vector))
                    symbols.append(chemical_symbols[labeling[count]])
                    count += 1
    structure = Atoms(symbols, positions, cell=new_cell, pbc=(True, True, True))
    structure.wrap()
    structure.pbc = pbc
    return structure


def get_symmetry_operations(structure: Atoms,
                            tolerance: float = 1e-3) -> Dict[str, list]:
    """
    Returns symmetry operations permissable for a given structure as
    obtained via `spglib <https://atztogo.github.io/spglib/>`_. The
    symmetry operations consist of three parts: rotation, translation
    and basis shifts. The latter define the way that sublattices
    shift upon rotation (correponds to `d_Nd` in [HarFor09]_).

    Parameters
    ----------
    structure
        structure for which symmetry operations are sought
    tolerance
        numerical tolerance imposed during symmetry analysis
    """

    symmetries = get_symmetry(structure)
    assert symmetries, ('spglib.get_symmetry() failed. Please make sure that'
                        ' the structure object is sensible.')
    rotations = symmetries['rotations']
    translations = symmetries['translations']

    basis = structure.get_scaled_positions()

    # Calculate how atoms within the primitive cell are shifted (from one site
    # to another) and translated (from one primtive cell to another) upon
    # operation with rotation matrix. Note that the translations are needed
    # here because different sites translate differently.
    basis_shifts = np.zeros((len(rotations), len(structure)), dtype='int64')
    sites_translations = []
    for i, rotation in enumerate(rotations):
        translation = translations[i]
        site_translations = []
        for j, basis_element in enumerate(basis):

            # Calculate how the site is transformed when operated on by
            # symmetry of parent lattice (rotation and translation)
            site_rot_trans = np.dot(rotation, basis_element) + translation

            # The site may now have been moved to a different site in a
            # different cell. We want to separate the two. (In HarFor09,
            # basis_shift (site_rot_trans) corresponds to d_Nd and
            # site_translation to t_Nd)
            site_translation = [0, 0, 0]
            for index in range(3):
                while site_rot_trans[index] < -tolerance:
                    site_rot_trans[index] += 1
                    site_translation[index] -= 1
                while site_rot_trans[index] > 1 - tolerance:
                    site_rot_trans[index] -= 1
                    site_translation[index] += 1
            site_translations.append(site_translation)

            # Find the basis element that the shifted basis correponds to
            found = False
            for basis_index, basis_element_comp in enumerate(basis):
                distance = site_rot_trans - basis_element_comp

                # Make sure that they do not differ with a basis vector
                for dist_comp_i, dist_comp in enumerate(distance):
                    if abs(abs(dist_comp) - 1) < tolerance:
                        distance[dist_comp_i] = 0

                if (abs(distance) < tolerance).all():
                    assert not found
                    basis_shifts[i, j] = basis_index
                    found = True
            assert found

        sites_translations.append(np.array(site_translations))

    symmetries['translations'] = sites_translations
    symmetries['basis_shifts'] = basis_shifts
    return symmetries


def enumerate_structures(structure: Atoms, sizes: List[int],
                         chemical_symbols: list,
                         concentration_restrictions: dict = None,
                         niggli_reduce: bool = None) -> Atoms:
    """
    Yields a sequence of enumerated structures. The function generates
    *all* inequivalent structures that are permissible given a certain
    lattice. Using the ``chemical_symbols`` and
    ``concentration_restrictions`` keyword arguments it is possible to
    specify which chemical_symbols are to be included on which site and
    in which concentration range.

    The function is sensitive to the boundary conditions of the input
    structure. An enumeration of, for example, a surface can thus be
    performed by setting ``structure.pbc = [True, True, False]``.

    The algorithm implemented here was developed by Gus L. W. Hart and
    Rodney W. Forcade in Phys. Rev. B **77**, 224115 (2008)
    [HarFor08]_ and Phys. Rev. B **80**, 014120 (2009) [HarFor09]_.

    Parameters
    ----------
    structure
        primitive structure from which derivative superstructures should
        be generated
    sizes
        number of sites (included in enumeration)
    chemical_symbols
        chemical species with which to decorate the structure, e.g.,
        ``['Au', 'Ag']``; see below for more examples
    concentration_restrictions
        allowed concentration range for one or more element in
        `chemical_symbols`, e.g., ``{'Au': (0, 0.2)}`` will only
        enumerate structures in which the Au content is between 0 and
        20 %; here, concentration is always defined as the number of
        atoms of the specified kind divided by the number of *all*
        atoms.
    niggli_reduction
        if True perform a Niggli reduction with spglib for each
        structure; the default is ``True`` if ``structure`` is periodic in
        all directions, ``False`` otherwise.

    Examples
    --------

    The following code snippet illustrates how to enumerate structures
    with up to 6 atoms in the unit cell for a binary alloy without any
    constraints::

        from ase.build import bulk
        prim = bulk('Ag')
        enumerate_structures(structure=prim, sizes=range(1, 7),
                             chemical_symbols=['Ag', 'Au'])

    To limit the concentration range to 10 to 40% Au the code should
    be modified as follows::

        enumerate_structures(structure=prim, sizes=range(1, 7),
                             chemical_symbols=['Ag', 'Au'],
                             concentration_restrictions={'Au': (0.1, 0.4)})

    Often one would like to consider mixing on only one
    sublattice. This can be achieved as illustrated for a
    Ga(1-x)Al(x)As alloy as follows::

        prim = bulk('GaAs', crystalstructure='zincblende', a=5.65)
        enumerate_structures(structure=prim, sizes=range(1, 9),
                             chemical_symbols=[['Ga', 'Al'], ['As']])

    """

    nsites = len(structure)
    basis = structure.get_scaled_positions()

    # Construct descriptor of where species are allowed to be
    if isinstance(chemical_symbols[0], str):
        iter_chemical_symbols = [tuple(range(len(chemical_symbols)))] * nsites
        elements = chemical_symbols
    elif len(chemical_symbols) == nsites:
        assert isinstance(chemical_symbols[0][0], str)
        elements = []
        for site in chemical_symbols:
            for element in site:
                if element not in elements:
                    elements.append(element)
        iter_chemical_symbols = []
        for site in chemical_symbols:
            iter_chemical_symbols.append(tuple(elements.index(i)
                                               for i in site))
    else:
        raise Exception('chemical_symbols needs to be a list of strings '
                        'or a list of list of strings.')

    # Adapt concentration restrictions to iter_chemical_symbols
    if concentration_restrictions:
        concentrations = {}
        for key, concentration_range in concentration_restrictions.items():
            assert len(concentration_range) == 2, \
                ('Each concentration range' +
                 ' needs to be specified as (c_low, c_high)')
            if key not in elements:
                raise ValueError('{} found in concentration_restrictions but'
                                 ' not in chemical_symbols'.format(key))
            concentrations[elements.index(key)] = concentration_range
    else:
        concentrations = None

    # Construct labeling generator
    labeling_generator = LabelingGenerator(iter_chemical_symbols,
                                           concentrations)

    # Niggli reduce by default if all directions have
    # periodic boundary conditions
    if niggli_reduce is None:
        niggli_reduce = (sum(structure.pbc) == 3)

    symmetries = get_symmetry_operations(structure)

    # Loop over each cell size
    for ncells in sizes:
        if ncells == 0:
            continue

        hnfs = list(yield_reduced_hnfs(ncells, symmetries, structure.pbc))
        snfs = get_unique_snfs(hnfs)

        for snf in snfs:
            labelings = _get_all_labelings(snf, labeling_generator, nsites)
            for hnf in snf.hnfs:
                if niggli_reduce:
                    new_cell = spg_nigg_red(np.dot(structure.cell.T, hnf.H).T)
                    if new_cell is None:
                        new_cell = np.dot(structure.cell.T, hnf.H).T
                else:
                    new_cell = np.dot(structure.cell.T, hnf.H).T
                for labeling in _yield_unique_labelings(labelings, snf, hnf,
                                                        nsites):
                    yield _labeling_to_structure(labeling, hnf, structure.cell,
                                                 new_cell, basis, elements,
                                                 structure.pbc)


def enumerate_supercells(structure: Atoms, sizes: List[int],
                         niggli_reduce: bool = None) -> Atoms:
    """
    Yields a sequence of enumerated supercells. The function generates
    *all* inequivalent supercells that are permissible given a certain
    lattice. Any supercell can be reduced to one of the supercells
    generated.

    The function is sensitive to the boundary conditions of the input
    structure. An enumeration of, for example, a surface can thus be
    performed by setting ``structure.pbc = [True, True, False]``.

    The algorithm is based on Gus L. W. Hart and
    Rodney W. Forcade in Phys. Rev. B **77**, 224115 (2008)
    [HarFor08]_ and Phys. Rev. B **80**, 014120 (2009) [HarFor09]_.

    Parameters
    ----------
    structure
        primitive structure from which supercells should be
        generated
    sizes
        number of sites (included in enumeration)
    niggli_reduction
        if True perform a Niggli reduction with spglib for each
        supercell; the default is ``True`` if ``structure`` is periodic in
        all directions, ``False`` otherwise.

    Examples
    --------

    The following code snippet illustrates how to enumerate supercells
    with up to 6 atoms in the unit cell::

        from ase.build import bulk
        prim = bulk('Ag')
        enumerate_supercells(structure=prim, sizes=range(1, 7))
    """

    # Niggli reduce by default if all directions have
    # periodic boundary conditions
    if niggli_reduce is None:
        niggli_reduce = (sum(structure.pbc) == 3)

    symmetries = get_symmetry_operations(structure)

    for ncells in sizes:
        for hnf in yield_reduced_hnfs(ncells, symmetries, structure.pbc):
            supercell = make_supercell(structure, hnf.H)
            if niggli_reduce:
                new_cell = spg_nigg_red(np.dot(structure.cell.T, hnf.H).T)
                if new_cell is None:  # Happens when spglib fails to Niggli reduce
                    yield supercell
                else:
                    Pprim = np.dot(new_cell, np.linalg.inv(structure.cell))
                    yield make_supercell(structure, Pprim)
            else:
                yield supercell



"""
Definition of the semi-grand canonical ensemble class.
"""

import numpy as np

from ase import Atoms
from ase.data import atomic_numbers, chemical_symbols
from ase.units import kB
from collections import OrderedDict
from typing import Dict, Union, List

from .. import DataContainer
from ..calculators.base_calculator import BaseCalculator
from .thermodynamic_base_ensemble import ThermodynamicBaseEnsemble


class SemiGrandCanonicalEnsemble(ThermodynamicBaseEnsemble):
    """Instances of this class allow one to simulate systems in the
    semi-grand canonical (SGC) ensemble (:math:`N\\Delta\\mu_i VT`), i.e. at
    constant temperature (:math:`T`), total number of sites (:math:`N=\\sum_i
    N_i`), relative chemical potentials (:math:`\\Delta\\mu_i=\\mu_i - \\mu_1`,
    where :math:`i` denotes the species), and volume (:math:`V`).

    The probability for a particular state in the SGC ensemble for a
    :math:`m`-component system can be written

    .. math::

        \\rho_{\\text{SGC}} \\propto \\exp\\Big[ - \\big( E
        + \\sum_{i>1}^m \\Delta\\mu_i N_i \\big) \\big / k_B T \\Big]

    with the *relative* chemical potentials :math:`\\Delta\\mu_i = \\mu_i -
    \\mu_1` and species counts :math:`N_i`. Unlike the :ref:`canonical ensemble
    <canonical_ensemble>`, the number of the respective species (or,
    equivalently, the concentrations) are allowed to vary in the SGC ensemble.
    A trial step thus consists of randomly picking an atom and changing its
    identity with probability

    .. math::

        P = \\min \\Big\\{ 1, \\, \\exp \\big[ - \\big( \\Delta E
        + \\sum_i \\Delta \\mu_i \\Delta N_i \\big) \\big / k_B T \\big]
        \\Big\\},

    where :math:`\\Delta E` is the change in potential energy caused by the
    swap.

    There exists a simple relation between the differences in chemical
    potential and the canonical free energy :math:`F`. In a binary system, this
    relationship reads

    .. math:: \\Delta \\mu = - \\frac{1}{N} \\frac{\\partial F}{\\partial c} (
        N, V, T, \\langle c \\rangle).

    Here :math:`c` denotes concentration (:math:`c=N_i/N`) and :math:`\\langle
    c \\rangle` the average concentration observed in the simulation. By
    recording :math:`\\langle c \\rangle` while gradually changing
    :math:`\\Delta \\mu`, one can thus in principle calculate the difference in
    canonical free energy between the pure phases (:math:`c=0` or :math:`1`)
    and any concentration by integrating :math:`\\Delta \\mu` over that
    concentration range. In practice this requires that the average recorded
    concentration :math:`\\langle c \\rangle` varies continuously with
    :math:`\\Delta \\mu`. This is not the case for materials with multiphase
    regions (such as miscibility gaps), because in such regions :math:`\\Delta
    \\mu` maps to multiple concentrations. In a Monte Carlo simulation, this is
    typically manifested by discontinuous jumps in concentration. Such jumps
    mark the phase boundaries of a multiphase region and can thus be used to
    construct the phase diagram. To recover the free energy, however, such
    systems require sampling in other ensembles, such as the
    :ref:`variance-constrained semi-grand canonical ensemble <sgc_ensemble>`.

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    temperature : float
        temperature :math:`T` in appropriate units [commonly Kelvin]
    chemical_potentials : Dict[str, float]
        chemical potential for each species :math:`\\mu_i`; the key
        denotes the species, the value specifies the chemical potential in
        units that are consistent with the underlying cluster expansion
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    sublattice_probabilities : List[float]
        probability for picking a sublattice when doing a random flip.
        This should be as long as the number of sublattices and should
        sum up to 1.

    Example
    -------
    The following snippet illustrate how to carry out a simple Monte Carlo
    simulation in the semi-canonical ensemble. Here, the parameters of the
    cluster expansion are set to emulate a simple Ising model in order to
    obtain an example that can be run without modification. In practice, one
    should of course use a proper cluster expansion::

        from ase.build import bulk
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import SemiGrandCanonicalEnsemble

        # prepare cluster expansion
        # the setup emulates a second nearest-neighbor (NN) Ising model
        # (zerolet and singlet ECIs are zero; only first and second neighbor
        # pairs are included)
        prim = bulk('Au')
        cs = ClusterSpace(prim, cutoffs=[4.3], chemical_symbols=['Ag', 'Au'])
        ce = ClusterExpansion(cs, [0, 0, 0.1, -0.02])

        # set up and run MC simulation (T=600 K, delta_mu=0.8 eV/atom)
        structure = prim.repeat(3)
        calc = ClusterExpansionCalculator(structure, ce)
        mc = SemiGrandCanonicalEnsemble(structure=structure, calculator=calc,
                                        temperature=600,
                                        data_container='myrun_sgc.dc',
                                        chemical_potentials={'Ag': 0, 'Au': 0.8})
        mc.run(100)  # carry out 100 trial swaps

    TODO
    ----
    * add check that chemical symbols in chemical potentials are allowed
    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 temperature: float,
                 chemical_potentials: Dict[str, float],
                 boltzmann_constant: float = kB,
                 user_tag: str = None,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None,
                 sublattice_probabilities: List[float] = None) -> None:

        self._ensemble_parameters = dict(temperature=temperature)

        # add chemical potentials to ensemble parameters
        self._chemical_potentials = get_chemical_potentials(chemical_potentials)
        for atnum, chempot in self.chemical_potentials.items():
            mu_sym = 'mu_{}'.format(chemical_symbols[atnum])
            self._ensemble_parameters[mu_sym] = chempot

        self._boltzmann_constant = boltzmann_constant

        super().__init__(
            structure=structure, calculator=calculator, user_tag=user_tag,
            random_seed=random_seed,
            data_container=data_container,
            data_container_class=DataContainer,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval,
            boltzmann_constant=boltzmann_constant)

        if sublattice_probabilities is None:
            self._flip_sublattice_probabilities = self._get_flip_sublattice_probabilities()
        else:
            self._flip_sublattice_probabilities = sublattice_probabilities

    @property
    def temperature(self) -> float:
        """ temperature :math:`T` (see parameters section above) """
        return self.ensemble_parameters['temperature']

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """
        sublattice_index = self.get_random_sublattice_index(
            probability_distribution=self._flip_sublattice_probabilities)
        return self.do_sgc_flip(self.chemical_potentials, sublattice_index=sublattice_index)

    @property
    def chemical_potentials(self) -> Dict[int, float]:
        """
        chemical potentials :math:`\\mu_i` (see parameters section above)
        """
        return self._chemical_potentials

    def _get_ensemble_data(self) -> Dict:
        """Returns the data associated with the ensemble. For the SGC
        ensemble this specifically includes the species counts.
        """
        # generic data
        data = super()._get_ensemble_data()

        # species counts
        data.update(self._get_species_counts())

        return data


def get_chemical_potentials(chemical_potentials: Dict[Union[int, str], float]):
    """ Gets values of chemical potentials."""
    if not isinstance(chemical_potentials, dict):
        raise TypeError('chemical_potentials has the wrong type: {}'
                        .format(type(chemical_potentials)))

    cps = OrderedDict([(key, val) if isinstance(key, int)
                       else (atomic_numbers[key], val)
                       for key, val in chemical_potentials.items()])

    return cps



"""
This module provides the ConvexHull class.
"""

import itertools
import numpy as np
from typing import List, Union
from scipy.interpolate import griddata
from scipy.spatial import ConvexHull as ConvexHullSciPy
from scipy.spatial.qhull import QhullError


class ConvexHull:
    """This class provides functionality for extracting the convex hull
    of the (free) energy of mixing. It is based on the `convex hull
    calculator in SciPy
    <http://docs.scipy.org/doc/scipy-dev/reference/\
generated/scipy.spatial.ConvexHull.html>`_.

    Parameters
    ----------
    concentrations : list(float) or list(list(float))
        concentrations for each structure listed as ``[[c1, c2], [c1, c2],
        ...]``; for binaries, in which case there is only one independent
        concentration, the format ``[c1, c2, c3, ...]`` works as well.
    energies : list(float)
        energy (or energy of mixing) for each structure

    Attributes
    ----------
    concentrations : np.ndarray
        concentrations of the `N` structures on the convex hull
    energies : np.ndarray
        energies of the `N` structures on the convex hull
    dimensions : int
        number of independent concentrations needed to specify a point in
        concentration space (1 for binaries, 2 for ternaries etc.)
    structures : list(int)
        indices of structures that constitute the convex hull (indices are
        defined by the order of their concentrations and energies are fed when
        initializing the ConvexHull object)

    Examples
    --------
    A `ConvexHull` object is easily initialized by providing lists of
    concentrations and energies::

        hull = ConvexHull(data['concentration'], data['mixing_energy'])

    after which one can for example plot the data (assuming a
    matplotlib axis object ``ax``)::

        ax.plot(hull.concentrations, hull.energies)

    or extract structures at or close to the convex hull::

        low_energy_structures = hull.extract_low_energy_structures(
            data['concentration'], data['mixing_energy'],
            energy_tolerance=0.005)

    A complete example can be found in the :ref:`basic tutorial
    <tutorial_enumerate_structures>`.

    """

    def __init__(self,
                 concentrations: Union[List[float], List[List[float]]],
                 energies: List[float]) -> None:
        assert len(concentrations) == len(energies)
        # Prepare data in format suitable for SciPy-ConvexHull
        concentrations = np.array(concentrations)
        energies = np.array(energies)
        points = np.column_stack((concentrations, energies))
        self.dimensions = len(points[0]) - 1

        # Construct convex hull
        hull = ConvexHullSciPy(points)

        # Collect convex hull points in handy arrays
        concentrations = []
        energies = []
        for vertex in hull.vertices:
            if self.dimensions == 1:
                concentrations.append(points[vertex][0])
            else:
                concentrations.append(points[vertex][0:-1])
            energies.append(points[vertex][-1])
        concentrations = np.array(concentrations)
        energies = np.array(energies)

        structures = hull.vertices
        # If there is just one independent concentration, we'd better sort
        # according to it
        if self.dimensions == 1:
            ces = list(zip(*sorted(zip(concentrations, energies, structures))))
            self.concentrations = np.array(ces[0])
            self.energies = np.array(ces[1])
            self.structures = np.array(ces[2])
        else:
            self.concentrations = concentrations
            self.energies = energies
            self.structures = structures

        # Remove points that are above the "pure components plane"
        self._remove_points_above_tie_plane()

    def _remove_points_above_tie_plane(self, tol: float = 1e-6) -> None:
        """
        Remove all points on the convex hull that correspond to maximum rather
        than minimum energy.

        Parameters
        ----------
        tol
            Tolerance for what energy constitutes a lower one.
        """

        # Identify the "complex concentration hull", i.e. the extremal
        # concentrations. In the simplest case, these should simply be the
        # pure components.
        if self.dimensions == 1:
            # Then the ConvexHullScipy function doesn't work, so we just pick
            # the indices of the lowest and highest concentrations.
            vertices = []
            vertices.append(np.argmin(self.concentrations))
            vertices.append(np.argmax(self.concentrations))
            vertices = np.array(vertices)
        else:
            concentration_hull = ConvexHullSciPy(self.concentrations)
            vertices = concentration_hull.vertices

        # Remove all points of the convex energy hull that have an energy that
        # is higher than what would be gotten with pure components at the same
        # concentration. These points are mathematically on the convex hull,
        # but in the physically uninteresting upper part, i.e. they maximize
        # rather than minimize energy.
        to_delete = []
        for i, concentration in enumerate(self.concentrations):
            # The points on the convex concentration hull should always be
            # included, so skip them.
            if i in vertices:
                continue

            # The energy obtained as a linear combination of concentrations on
            # the convex hull is the "z coordinate" of the position on a
            # (hyper)plane in the (number of independent concentrations +
            # 1)-dimensional (N-D) space. This plane is spanned by N points.
            # If there are more vertices on the convex hull, we need to loop
            # over all combinations of N vertices.
            for plane in itertools.combinations(vertices,
                                                min(len(vertices),
                                                    self.dimensions + 1)):
                # Calculate energy that would be gotten with pure components
                # with ascribed concentration.
                energy_pure = griddata(self.concentrations[np.array(plane)],
                                       self.energies[np.array(plane)],
                                       concentration,
                                       method='linear')

                # Prepare to delete if the energy was lowered. `griddata` gives
                # NaN if the concentration is outside the triangle formed by
                # the three vertices. The result of the below comparison is
                # then False, which is what we want.
                if energy_pure < self.energies[i] - tol:
                    to_delete.append(i)
                    break

        # Finally remove all points
        self.concentrations = np.delete(self.concentrations, to_delete, 0)
        self.energies = np.delete(self.energies, to_delete, 0)
        self.structures = list(np.delete(self.structures, to_delete, 0))

    def get_energy_at_convex_hull(self, target_concentrations:
                                  Union[List[float],
                                        List[List[float]]]) -> np.ndarray:
        """Returns the energy of the convex hull at specified concentrations.
        If any concentration is outside the allowed range, NaN is
        returned.

        Parameters
        ----------
        target_concentrations
            concentrations at target points

            If there is one independent concentration, a list of
            floats is sufficient. Otherwise, the concentrations ought
            to be provided as a list of lists, such as ``[[0.1, 0.2],
            [0.3, 0.1], ...]``.
        """
        if self.dimensions > 1:
            assert len(target_concentrations[0]) == self.dimensions

        # Loop over all complexes of N+1 points to make sure that the lowest
        # energy plane is used in the end. This is needed in two dimensions
        # but in higher.
        hull_candidate_energies = []
        for plane in itertools.combinations(range(len(self.energies)),
                                            min(len(self.energies),
                                                self.dimensions + 1)):
            try:
                plane_energies = griddata(self.concentrations[list(plane)],
                                          self.energies[list(plane)],
                                          np.array(target_concentrations),
                                          method='linear')
            except QhullError:
                # If the points lie on a line, the convex hull will fail, but
                # we do not need to care about these "planes" anyway
                continue
            hull_candidate_energies.append(plane_energies)

        # Pick out the lowest energies found
        hull_energies = np.nanmin(hull_candidate_energies, axis=0)
        return hull_energies

    def extract_low_energy_structures(self, concentrations:
                                      Union[List[float],
                                            List[List[float]]],
                                      energies: List[float],
                                      energy_tolerance: float) -> List[int]:
        """Returns the indices of energies that lie within a certain
        tolerance of the convex hull.

        Parameters
        ----------
        concentrations
            concentrations of candidate structures

            If there is one independent concentration, a list of
            floats is sufficient. Otherwise, the concentrations must
            be provided as a list of lists, such as ``[[0.1, 0.2],
            [0.3, 0.1], ...]``.
        energies
            energies of candidate structures
        energy_tolerance
            include structures with an energy that is at most this far
            from the convex hull
        """
        # Convert to numpy arrays, can be necessary if, for example,
        # they are Pandas Series with "gaps"
        concentrations = np.array(concentrations)
        energies = np.array(energies)

        n_points = len(concentrations)
        if len(energies) != n_points:
            raise ValueError('concentrations and energies must have '
                             'the same length')

        # Calculate energy at convex hull for specified concentrations
        hull_energies = self.get_energy_at_convex_hull(concentrations)

        # Extract those that are close enough
        close_to_hull = [i for i in range(n_points)
                         if energies[i] <= hull_energies[i] + energy_tolerance]

        return close_to_hull



"""Definition of the abstract base ensemble class."""

import os
import random
from abc import ABC, abstractmethod
from collections import OrderedDict
from math import gcd
from time import time
from typing import BinaryIO, Dict, List, TextIO, Union

import numpy as np

from ase import Atoms
from icet.core.sublattices import Sublattices

from ..calculators.base_calculator import BaseCalculator
from ..configuration_manager import ConfigurationManager
from ..data_containers.base_data_container import BaseDataContainer
from ..observers.base_observer import BaseObserver


class BaseEnsemble(ABC):
    """Base ensemble class.

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    data_container_class : BaseDataContainer
        used to initialize custom (ensemble specific) data container objects;
        by default the class uses the generic BaseDataContainer class
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data.
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 user_tag: str = None,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_class: BaseDataContainer = BaseDataContainer,
                 data_container_write_period: float = 600,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None) -> None:

        # initialize basic variables
        self._accepted_trials = 0
        self._observers = {}
        self._step = 0

        # calculator and configuration
        self._calculator = calculator
        self._user_tag = user_tag
        sublattices = self.calculator.sublattices

        sublattices.assert_occupation_is_allowed(structure.get_chemical_symbols())

        # item for sublist in l for item in sublist
        symbols_flat = [s for sl in sublattices.active_sublattices for s in sl.chemical_symbols]
        if len(symbols_flat) != len(set(symbols_flat)):
            bad_symbols = set([s for s in symbols_flat if symbols_flat.count(s) > 1])
            raise ValueError('Symbols {} found on multiple active sublattices'.format(bad_symbols))

        self.configuration = ConfigurationManager(structure, sublattices)

        # random number generator
        if random_seed is None:
            self._random_seed = random.randint(0, 1e16)
        else:
            self._random_seed = random_seed
        random.seed(a=self._random_seed)

        # add ensemble parameters and metadata
        self._ensemble_parameters['n_atoms'] = len(self.structure)
        metadata = OrderedDict(ensemble_name=self.__class__.__name__,
                               user_tag=user_tag, seed=self.random_seed)

        # data container
        self._data_container_write_period = data_container_write_period
        self._data_container_filename = data_container

        if data_container is not None and os.path.isfile(data_container):
            self._data_container = BaseDataContainer.read(data_container)

            dc_ensemble_parameters = self.data_container.ensemble_parameters
            if not dicts_equal(self.ensemble_parameters,
                               dc_ensemble_parameters):
                raise ValueError('Ensemble parameters do not match those'
                                 ' stored in data container file: {}'.format(
                                     set(dc_ensemble_parameters.items()) -
                                     set(self.ensemble_parameters.items())))
            self._restart_ensemble()
        else:
            if data_container is not None:
                # check if path to file exists
                filedir = os.path.dirname(data_container)
                if filedir and not os.path.isdir(filedir):
                    raise FileNotFoundError('Path to data container file does'
                                            ' not exist: {}'.format(filedir))
            self._data_container = data_container_class(
                structure=structure,
                ensemble_parameters=self.ensemble_parameters,
                metadata=metadata)

        # interval for writing data and further preparation of data container
        self._default_interval = len(structure)

        if ensemble_data_write_interval is None:
            self._ensemble_data_write_interval = self._default_interval
        else:
            self._ensemble_data_write_interval = ensemble_data_write_interval

        # Handle trajectory writing
        if trajectory_write_interval is None:
            self._trajectory_write_interval = self._default_interval
        else:
            self._trajectory_write_interval = trajectory_write_interval

        self._find_observer_interval()

    @property
    def structure(self) -> Atoms:
        """ current configuration (copy) """
        return self.configuration.structure.copy()

    @property
    def data_container(self) -> BaseDataContainer:
        """ data container associated with ensemble """
        return self._data_container

    @property
    def observers(self) -> Dict[str, BaseObserver]:
        """ observers """
        return self._observers

    @property
    def calculator(self) -> BaseCalculator:
        """ calculator attached to the ensemble """
        return self._calculator

    @property
    def step(self) -> int:
        """ current configuration (copy) """
        return self._step

    def run(self, number_of_trial_steps: int):
        """
        Samples the ensemble for the given number of trial steps.

        Parameters
        ----------
        number_of_trial_steps
            number of MC trial steps to run in total
        reset_step
            if True the MC trial step counter and the data container will
            be reset to zero and empty, respectively.

        Raises
        ------
        TypeError
            if `number_of_trial_steps` is not an int
        """

        if not isinstance(number_of_trial_steps, int):
            raise TypeError('number_of_trial_steps must be an integer ({})'
                            .format(number_of_trial_steps))

        last_write_time = time()

        initial_step = self.step
        final_step = self.step + number_of_trial_steps
        # run Monte Carlo simulation such that we start at an
        # interval which lands on the observer interval
        if initial_step != 0:
            first_run_interval = self.observer_interval -\
                (initial_step -
                 (initial_step // self.observer_interval) *
                 self.observer_interval)
            first_run_interval = min(first_run_interval, number_of_trial_steps)
            self._run(first_run_interval)
            initial_step += first_run_interval

        step = initial_step
        while step < final_step and not self._terminate_sampling():
            uninterrupted_steps = min(self.observer_interval, final_step - step)
            if self.step % self.observer_interval == 0:
                self._observe(self.step)
            if self._data_container_filename is not None and \
                    time() - last_write_time > self._data_container_write_period:
                self.write_data_container(self._data_container_filename)
                last_write_time = time()

            self._run(uninterrupted_steps)
            step += uninterrupted_steps

        # if we end on an observation interval we also observe
        if self.step % self.observer_interval == 0:
            self._observe(self.step)

        # allow ensemble a chance to go clean
        self._finalize()

        if self._data_container_filename is not None:
            self.write_data_container(self._data_container_filename)

    def _run(self, number_of_trial_steps: int):
        """Runs MC simulation for a number of trial steps without
        interruption.

        Parameters
        ----------
        number_of_trial_steps
            number of trial steps to run without stopping
        """
        for _ in range(number_of_trial_steps):
            accepted = self._do_trial_step()
            self._step += 1
            self._accepted_trials += accepted

    def _observe(self, step: int):
        """Submits current configuration to observers and appends
        observations to data container.

        Parameters
        ----------
        step
            the current trial step
        """
        row_dict = {}

        # Ensemble specific data
        if step % self._ensemble_data_write_interval == 0:
            ensemble_data = self._get_ensemble_data()
            for key, value in ensemble_data.items():
                row_dict[key] = value

            # reset accepted trial count
            self._accepted_trials = 0

        # Trajectory data
        if step % self._trajectory_write_interval == 0:
            row_dict['occupations'] = self.configuration.occupations.tolist()

        # Observer data
        for observer in self.observers.values():
            if step % observer.interval == 0:
                if observer.return_type is dict:
                    for key, value in observer.get_observable(self.calculator.structure).items():
                        row_dict[key] = value
                else:
                    row_dict[observer.tag] = observer.get_observable(self.calculator.structure)

        if len(row_dict) > 0:
            self._data_container.append(mctrial=step, record=row_dict)

    @abstractmethod
    def _do_trial_step(self):
        pass

    @property
    def user_tag(self) -> str:
        """ tag used for labeling the ensemble """
        return self._user_tag

    @property
    def random_seed(self) -> int:
        """ seed used to initialize random number generator """
        return self._random_seed

    def _next_random_number(self) -> float:
        """ Returns the next random number from the PRNG. """
        return random.random()

    @property
    def observer_interval(self) -> int:
        """minimum number of steps to run Monte Carlo simulation without
        interruption for observation
        """
        return self._observer_interval

    def _find_observer_interval(self) -> int:
        """
        Finds the greatest common denominator from the observation intervals.
        """
        intervals = [obs.interval for obs in self.observers.values()]

        if self._ensemble_data_write_interval is not np.inf:
            intervals.append(self._ensemble_data_write_interval)
        if self._trajectory_write_interval is not np.inf:
            intervals.append(self._trajectory_write_interval)
        if intervals:
            self._observer_interval = self._get_gcd(intervals)

    def _get_gcd(self, values: List[int]) -> int:
        """ Finds the greatest common denominator (GCD) from a list of integers. """
        if len(values) == 1:
            return values[0]

        if len(values) > 2:
            gcd_right = gcd(values[-1], values[-2])
            values.pop()
            values.pop()
            values.append(gcd_right)
            return self._get_gcd(values)
        else:
            return gcd(values[0], values[1])

    def attach_observer(self, observer: BaseObserver, tag=None):
        """
        Attaches an observer to the ensemble.

        Parameters
        ----------
        observer
            observer instance to attach
        tag
            name used in data container
        """
        if not isinstance(observer, BaseObserver):
            raise TypeError('observer has the wrong type: {}'.format(type(observer)))

        if observer.interval is None:
            observer.interval = self._default_interval

        if tag is not None:
            observer.tag = tag
            self.observers[tag] = observer
        else:
            self.observers[observer.tag] = observer

        self._find_observer_interval()

    def update_occupations(self, sites: List[int], species: List[int]):
        """Updates the occupation vector of the configuration being
        sampled. This will change the state of the configuration in
        both the calculator and the configuration manager.

        Parameters
        ----------
        sites
            indices of sites of the configuration to change
        species
            new occupations (species) by atomic number

        Raises
        ------
        ValueError
            if input lists are not of the same length
        """

        if len(sites) != len(species):
            raise ValueError('sites and species must have the same length.')
        self.configuration.update_occupations(sites, species)

    def _get_property_change(self, sites: List[int], species: List[int]) -> float:
        """Computes and returns the property change due to a change of
        the configuration.

        _N.B.:_ This method leaves the configuration itself unchanged.

        Parameters
        ----------
        sites
            indices of sites to change
        species
            new occupations (species) by atomic number
        """
        current_species = self.configuration.occupations[sites]
        current_property = self.calculator.calculate_local_contribution(
            local_indices=sites,
            occupations=self.configuration.occupations)

        self.update_occupations(sites=sites, species=species)
        new_property = self.calculator.calculate_local_contribution(
            local_indices=sites,
            occupations=self.configuration.occupations)
        property_change = new_property - current_property

        # Restore initial configuration
        self.update_occupations(sites, current_species)
        return property_change

    def _get_ensemble_data(self) -> dict:
        """ Returns the current calculator property. """
        potential = self.calculator.calculate_total(occupations=self.configuration.occupations)
        return {'potential': potential,
                'acceptance_ratio': self._accepted_trials / self._ensemble_data_write_interval}

    def get_random_sublattice_index(self, probability_distribution) -> int:
        """Returns a random sublattice index based on the weights of the
        sublattice.

        Parameters
        ----------
        probability_distribution
            probability distributions for the sublattices
        """

        if len(probability_distribution) != len(self.sublattices):
            raise ValueError('probability_distribution should have the same size as sublattices')
        pick = np.random.choice(len(self.sublattices), p=probability_distribution)
        return pick

    def _restart_ensemble(self):
        """Restarts ensemble using the last state saved in data container file.
        """

        # Restart step
        self._step = self.data_container.last_state['last_step']

        # Update configuration
        occupations = self.data_container.last_state['occupations']
        active_sites = []
        for sl in self.sublattices.active_sublattices:
            active_sites.extend(sl.indices)
        active_occupations = [occupations[s] for s in active_sites]
        self.update_occupations(active_sites, active_occupations)

        # Restart number of total and accepted trial steps
        self._accepted_trials = self.data_container.last_state['accepted_trials']

        # Restart state of random number generator
        random.setstate(self.data_container.last_state['random_state'])

    def write_data_container(self, outfile: Union[str, BinaryIO, TextIO]):
        """Updates last state of the Monte Carlo simulation and
        writes data container to file.

        Parameters
        ----------
        outfile
            file to which to write
        """
        self._data_container._update_last_state(
            last_step=self.step,
            occupations=self.configuration.occupations.tolist(),
            accepted_trials=self._accepted_trials,
            random_state=random.getstate())

        self.data_container.write(outfile)

    @property
    def ensemble_parameters(self) -> dict:
        """Returns parameters associated with the ensemble."""
        return self._ensemble_parameters.copy()

    @property
    def sublattices(self) -> Sublattices:
        """sublattices for the configuration being sampled"""
        return self.configuration.sublattices

    def _terminate_sampling(self) -> bool:
        """This method is called from the run method to determine whether the MC
        sampling loop should be terminated for a reason other than having exhausted
        the number of iterations. The method can be overriden by child classes in
        order to provide an alternative exit mechanism.
        """
        return False

    def _finalize(self) -> None:
        """This method is called from the run method after the conclusion of
        the MC cycles but before the data container is written. This
        method can be used by child classes to carry out clean-up
        tasks, including e.g., adding "left-over" data to the data
        container.
        """
        pass


def dicts_equal(dict1: Dict, dict2: Dict, atol: float = 1e-12) -> bool:

    """Returns True (False) if two dicts are equal (not equal), if
    float or integers are in the dicts then atol is used for comparing them."""
    if len(dict1) != len(dict2):
        return False
    for key in dict1.keys():
        if key not in dict2:
            return False
        if isinstance(dict1[key], (int, float)) and isinstance(dict2[key], (int, float)):
            if not np.isclose(dict1[key], dict2[key], rtol=0.0, atol=atol) and \
                   not np.isnan(dict1[key]) and not np.isnan(dict2[key]):
                return False
        else:
            if dict1[key] != dict2[key]:
                return False
    return True



from typing import Dict, List

import pandas as pd

from _icet import ClusterCounts as _ClusterCounts
from ase import Atoms
from icet.core.cluster import Cluster
from icet.core.local_orbit_list_generator import LocalOrbitListGenerator
from icet.core.structure import Structure
from mchammer.observers.base_observer import BaseObserver


class ClusterCountObserver(BaseObserver):
    """
    This class represents a cluster count observer.

    A cluster count observer enables one to keep track of the
    occupation of clusters along the trajectory sampled by a Monte
    Carlo (MC) simulation. For example, using this observer, several
    canonical MC simulations could be carried out at different
    temperatures and the temperature dependence of the number of
    nearest neigbhors of a particular species could be accessed with
    this observer.

    Parameters
    ----------
    cluster_space : icet.ClusterSpace
     cluster space to define the clusters to be counted
    structure : ase.Atoms
        defines the lattice that the observer will work on
    interval : int
        observation interval during the Monte Carlo simulation

    Attributes
    ----------
    tag : str
        human readable observer name
    interval : int
        observation interval
    """

    def __init__(self, cluster_space, structure: Atoms,
                 interval: int) -> None:
        super().__init__(interval=interval, return_type=dict, tag='ClusterCountObserver')

        self._cluster_space = cluster_space
        local_orbit_list_generator = LocalOrbitListGenerator(cluster_space.orbit_list,
                                                             Structure.from_atoms(structure))

        self._full_orbit_list = local_orbit_list_generator.generate_full_orbit_list()
        self._cluster_counts_cpp = _ClusterCounts()

        self._cluster_keys = []
        for i, orbit in enumerate(self._full_orbit_list.orbits):
            cluster = orbit.representative_cluster
            cluster.tag = i
            self._cluster_keys.append(cluster)

        self._get_empty_counts()

    def _get_empty_counts(self) -> Dict[Cluster, Dict[List[str], int]]:
        """ Returns the object which will be filled with counts. """
        counts = {}
        for i, cluster in enumerate(self._cluster_keys):
            order = len(cluster)
            possible_occupations = self._cluster_space.get_possible_orbit_occupations(cluster.tag)
            assert order == len(possible_occupations[0]), '{} is not {}, {}, {}'.format(
                order, len(possible_occupations[0]), possible_occupations)

            counts[cluster] = {occupation: 0 for occupation in possible_occupations}
        return counts

    def _generate_counts(self, structure: Atoms) -> None:
        """Counts the occurrence of different clusters and stores this
        information in a pandas dataframe.

        Parameters
        ----------
        structure
            input atomic structure.
        """
        self._cluster_counts_cpp.count_orbit_list(Structure.from_atoms(structure),
                                                  self._full_orbit_list, True, True)

        empty_counts = self._get_empty_counts()
        pandas_rows = []

        # std::unordered_map<Cluster, std::map<std::vector<int>, int>>
        cluster_counts = self._cluster_counts_cpp.get_cluster_counts()

        for cluster_key, chemical_number_counts_dict in cluster_counts.items():

            for chemical_symbols in empty_counts[cluster_key].keys():

                count = chemical_number_counts_dict.get(chemical_symbols, 0)
                pandas_row = {}
                pandas_row['dc_tag'] = '{}_{}'.format(cluster_key.tag, '_'.join(chemical_symbols))
                pandas_row['occupation'] = chemical_symbols
                pandas_row['cluster_count'] = count
                pandas_row['orbit_index'] = cluster_key.tag
                pandas_row['order'] = len(cluster_key)
                pandas_row['radius'] = cluster_key.radius
                pandas_rows.append(pandas_row)
        self.count_frame = pd.DataFrame(pandas_rows)
        self._cluster_counts_cpp.reset()

    def get_observable(self, structure: Atoms) -> dict:
        """
        Returns the value of the property from a cluster expansion model
        for a given atomic configuration.

        Parameters
        ----------
        structure
            input atomic structure
        """
        self._generate_counts(structure)

        count_dict = {row['dc_tag']: row['cluster_count']
                      for i, row in self.count_frame.iterrows()}
        return count_dict



import numpy as np
import spglib
from typing import List, Sequence, Tuple, TypeVar
from ase import Atoms
from ase.data import chemical_symbols
from icet.core.lattice_site import LatticeSite
from icet.core.neighbor_list import NeighborList
from icet.core.structure import Structure

Vector = List[float]
T = TypeVar('T')


def get_scaled_positions(positions: np.ndarray,
                         cell: np.ndarray,
                         wrap: bool = True,
                         pbc: List[bool] = [True, True, True]) \
        -> np.ndarray:
    """
    Returns the positions in reduced (or scaled) coordinates.

    Parameters
    ----------
    positions
        atomic positions in cartesian coordinates
    cell
        cell metric
    wrap
        if True, positions outside the unit cell will be wrapped into
        the cell in the directions with periodic boundary conditions
        such that the scaled coordinates are between zero and one.
    pbc
        periodic boundary conditions flags
    """

    fractional = np.linalg.solve(cell.T, positions.T).T

    if wrap:
        for i, periodic in enumerate(pbc):
            if periodic:
                # Yes, we need to do it twice.
                # See the scaled_positions.py test.
                fractional[:, i] %= 1.0
                fractional[:, i] %= 1.0

    return fractional


def get_primitive_structure(structure: Atoms,
                            no_idealize: bool = True,
                            to_primitive: bool = True,
                            symprec: float = 1e-5) -> Atoms:
    """
    Returns the primitive structure using spglib.

    Parameters
    ----------
    structure
        input atomic structure
    no_idealize
        if True lengths and angles are not idealized
    to_primitive
        convert to primitive structure
    symprec
        tolerance imposed when analyzing the symmetry using spglib
    """
    structure_cpy = structure.copy()
    structure_as_tuple = ase_atoms_to_spglib_cell(structure_cpy)

    lattice, scaled_positions, numbers = spglib.standardize_cell(
        structure_as_tuple, to_primitive=to_primitive,
        no_idealize=no_idealize, symprec=symprec)
    scaled_positions = [np.round(pos, 12) for pos in scaled_positions]
    structure_prim = Atoms(scaled_positions=scaled_positions,
                           numbers=numbers, cell=lattice, pbc=structure.pbc)
    structure_prim.wrap()

    return structure_prim


def get_fractional_positions_from_neighbor_list(
        structure: Structure, neighbor_list: NeighborList) -> List[Vector]:
    """
    Returns the fractional positions of the lattice sites in structure from
    a neighbor list.

    Parameters
    ----------
    structure
        input atomic structure
    neighbor_list
        list of lattice neighbors of the input structure
    """
    neighbor_positions = []
    fractional_positions = []
    lattice_site = LatticeSite(0, [0, 0, 0])

    for i in range(len(neighbor_list)):
        lattice_site.index = i
        position = structure.get_position(lattice_site)
        neighbor_positions.append(position)
        for neighbor in neighbor_list.get_neighbors(i):
            position = structure.get_position(neighbor)
            neighbor_positions.append(position)

    if len(neighbor_positions) > 0:
        fractional_positions = get_scaled_positions(
            np.array(neighbor_positions),
            structure.cell, wrap=False,
            pbc=structure.pbc)

    return fractional_positions


def get_position_from_lattice_site(structure: Atoms, lattice_site: LatticeSite):
    """
    Gets the corresponding position from the lattice site.

    Parameters
    ---------
    structure
        input atomic structure
    lattice_site
        specific lattice site of the input structure
    """
    return structure[lattice_site.index].position + \
        np.dot(lattice_site.unitcell_offset, structure.get_cell())


def find_lattice_site_by_position(structure: Atoms, position: List[float],
                                  tol: float = 1e-4) -> LatticeSite:
    """
    Tries to construct a lattice site equivalent from
    position in reference to the ASE Atoms object.

    structure
        input atomic structure
    position
        presumed cartesian coordinates of a lattice site
    """

    for i, atom in enumerate(structure):
        pos = position - atom.position
        # Direct match
        if np.linalg.norm(pos) < tol:
            return LatticeSite(i, np.array((0., 0., 0.)))

        fractional = np.linalg.solve(structure.cell.T, np.array(pos).T).T
        unit_cell_offset = [np.floor(round(x)) for x in fractional]
        residual = np.dot(fractional - unit_cell_offset, structure.cell)
        if np.linalg.norm(residual) < tol:
            latNbr = LatticeSite(i, unit_cell_offset)
            return latNbr

    # found nothing, raise error
    raise RuntimeError('Did not find site in find_lattice_site_by_position')


def fractional_to_cartesian(structure: Atoms,
                            frac_positions: List[Vector]) -> List[Vector]:
    """
    Turns fractional positions into cartesian positions.

    Parameters
    ----------
    structure
        input atomic structure
    frac_positions
        fractional positions
    """
    return np.dot(frac_positions, structure.cell)


def get_permutation(container: Sequence[T],
                    permutation: List[int]) -> Sequence[T]:
    """
    Returns the permuted version of container.
    """
    if len(permutation) != len(container):
        raise RuntimeError('Container and permutation'
                           ' not of same size {} != {}'.format(
                               len(container), len(permutation)))
    if len(set(permutation)) != len(permutation):
        raise Exception
    return [container[s] for s in permutation]


def ase_atoms_to_spglib_cell(structure: Atoms) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Returns a tuple of three components: cell metric, atomic positions, and
    atomic species of the input ASE Atoms object.
    """
    return (structure.get_cell(), structure.get_scaled_positions(), structure.get_atomic_numbers())


def get_occupied_primitive_structure(
        structure: Atoms,
        allowed_species: List[List[str]],
        symprec: float = 1e-5) -> Tuple[Atoms, List[List[str]]]:
    """
    Returns an occupied primitive structure.
    Will put hydrogen on sublattice 1, Helium on sublattice 2 and
    so on

    Parameters
    ----------
    structure
        input structure
    allowed_species
        chemical symbols that are allowed on each site
    symprec
        tolerance imposed when analyzing the symmetry using spglib

    Todo
    ----
    simplify the revert back to unsorted symbols
    """
    if len(structure) != len(allowed_species):
        raise ValueError(
            'structure and chemical symbols need to be the same size.')
    symbols = sorted({tuple(sorted(s)) for s in allowed_species})

    decorated_primitive = structure.copy()
    for i, sym in enumerate(allowed_species):
        sublattice = symbols.index(tuple(sorted(sym))) + 1
        decorated_primitive[i].symbol = chemical_symbols[sublattice]

    decorated_primitive = get_primitive_structure(decorated_primitive,
                                                  symprec=symprec)
    decorated_primitive.wrap()
    primitive_chemical_symbols = []
    for atom in decorated_primitive:
        sublattice = chemical_symbols.index(atom.symbol)
        primitive_chemical_symbols.append(symbols[sublattice - 1])

    for symbols in allowed_species:
        if tuple(sorted(symbols)) in primitive_chemical_symbols:
            index = primitive_chemical_symbols.index(tuple(sorted(symbols)))
            primitive_chemical_symbols[index] = symbols
    return decorated_primitive, primitive_chemical_symbols


def atomic_number_to_chemical_symbol(numbers: List[int]) -> List[str]:
    """Returns the chemical symbols equivalent to the input atomic
    numbers.

    Parameters
    ----------
    numbers
        atomic numbers
    """

    symbols = [chemical_symbols[number] for number in numbers]
    return symbols


def chemical_symbols_to_numbers(symbols: List[str]) -> List[int]:
    """Returns the atomic numbers equivalent to the input chemical
    symbols.

    Parameters
    ----------
    symbols
        chemical symbols
    """

    numbers = [chemical_symbols.index(symbols) for symbols in symbols]
    return numbers


def get_wyckoff_sites(structure: Atoms, symprec: float = 1e-4,
                      map_occupations: List[List[str]] = None) -> List[str]:
    """Returns the Wyckoff symbols of the input structure. The Wyckoff
    sites are of general interest for symmetry analysis but can be
    especially useful when setting up, e.g., a
    :class:`SiteOccupancyObserver
    <mchammer.observers.SiteOccupancyObserver>`.
    The Wyckoff labels can be conveniently attached as an array to the
    structure object as demonstrated in the Examples section below.

    By default the occupation of the sites is part of the symmetry
    analysis. If a chemically disordered structure is provided this
    will usually reduce the symmetry substantially. If one is
    interested in the symmetry of the underlying structure one can
    control how occupations are handled. To this end, one can provide
    the ``map_occupations`` keyword argument. The latter must be a
    list, each entry of which is a list of species that should be
    treated as indistinguishable. As a shortcut, if *all* species
    should be treated as indistinguishable one can provide an empty
    list. Examples that illustrate the usage of the keyword are given
    below.

    Parameters
    ----------
    structure
        input structure, note that the occupation of the sites is
        included in the symmetry analysis
    symprec
        tolerance parameter handed over to spglib
    map_occupations
        each sublist in this list specifies a group of chemical
        species that shall be treated as indistinguishable for the
        purpose of the symmetry analysis

    Examples
    --------
    Wyckoff sites of a hexagonal-close packed structure::

        >> from ase.build import bulk
        >> from icet.tools import get_wyckoff_sites

        >> structure = bulk('Ti')
        >> wyckoff_sites = get_wyckoff_sites(structure)
        >> print(wyckoff_sites)

    Running the snippet above will produce the following output::

        ['2d', '2d']

    The Wyckoff labels can also be attached as an array to the
    structure, in which case the information is also included when
    storing the Atoms object::

        >> from ase.io import write
        >> structure.new_array('wyckoff_sites', wyckoff_sites, str)
        >> write('structure.xyz', structure)

    The function can also be applied to supercells::

        >> structure = bulk('GaAs', crystalstructure='zincblende', a=3.0).repeat(2)
        >> wyckoff_sites = get_wyckoff_sites(structure)
        >> print(wyckoff_sites)

    This snippet will produce the following output::

        ['4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c',
         '4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c']

    Now assume that one is given a supercell of a (Ga,Al)As
    alloy. Applying the function directly yields much lower symmetry
    since the symmetry of the original structure is broken::

        >> structure.set_chemical_symbols(
               ['Ga', 'As', 'Al', 'As', 'Ga', 'As', 'Al', 'As',
                'Ga', 'As', 'Ga', 'As', 'Al', 'As', 'Ga', 'As'])
        >> print(get_wyckoff_sites(structure))

        ['8g', '8i', '4e', '8i', '8g', '8i', '2c', '8i',
         '2d', '8i', '8g', '8i', '4e', '8i', '8g', '8i']

    Since Ga and Al occupy the same sublattice, they should, however,
    be treated as indistinguishable for the purpose of the symmetry
    analysis, which can be achieved via the ``map_occupations``
    keyword::

        >> print(get_wyckoff_sites(structure, map_occupations=[['Ga', 'Al'], ['As']]))

        ['4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c',
         '4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c']

    If occupations are to ignored entirely, one can simply provide an
    empty list. In the present case, this turns the zincblende lattice
    into a diamond lattice, on which case there is only one Wyckoff
    site::

        >> print(get_wyckoff_sites(structure, map_occupations=[]))

        ['8a', '8a', '8a', '8a', '8a', '8a', '8a', '8a',
         '8a', '8a', '8a', '8a', '8a', '8a', '8a', '8a']
    """
    structure_copy = structure.copy()
    if map_occupations is not None:
        if len(map_occupations) > 0:
            new_symbols = []
            for symb in structure_copy.get_chemical_symbols():
                for group in map_occupations:
                    if symb in group:
                        new_symbols.append(group[0])
                        break
        else:
            new_symbols = len(structure) * ['H']
        structure_copy.set_chemical_symbols(new_symbols)
    dataset = spglib.get_symmetry_dataset((structure_copy.get_cell(),
                                           structure_copy.get_scaled_positions(),
                                           structure_copy.get_atomic_numbers()),
                                          symprec=symprec)
    n_unitcells = np.linalg.det(dataset['transformation_matrix'])

    equivalent_atoms = list(dataset['equivalent_atoms'])
    wyckoffs = {}
    for index in set(equivalent_atoms):
        multiplicity = list(dataset['equivalent_atoms']).count(index) / n_unitcells
        multiplicity = int(round(multiplicity))
        wyckoffs[index] = '{}{}'.format(multiplicity, dataset['wyckoffs'][index])

    return [wyckoffs[equivalent_atoms[a.index]] for a in structure_copy]



from ase import Atoms
from icet import ClusterExpansion
from mchammer.observers.base_observer import BaseObserver


class ClusterExpansionObserver(BaseObserver):
    """
    This class represents a cluster expansion (CE) observer.

    A CE observer allows to compute a property described by a CE along the
    trajectory sampled by a Monte Carlo (MC) simulation. In general this CE
    differs from the CE that is used to generate the trajectory. For example in
    a canonical MC simulation the latter would usually represent an energy
    (total or mixing energy) whereas the former CE(s) could map lattice
    constant or band gap.

    Parameters
    ----------
    cluster_expansion : :class:`icet.ClusterExpansion` cluster expansion model
        to be used for observation
    tag : str
        human readable observer name (default: `ClusterExpansionObserver`)
    interval : int
        observation interval during the Monte Carlo simulation

    Attributes
    ----------
    tag : str
        name of observer
    interval : int
        observation interval
    """

    def __init__(self, cluster_expansion: ClusterExpansion,
                 interval: int = None,
                 tag: str = 'ClusterExpansionObserver') -> None:
        super().__init__(interval=interval, return_type=float, tag=tag)
        self._cluster_expansion = cluster_expansion

    def get_observable(self, structure: Atoms) -> float:
        """
        Returns the value of the property from a cluster expansion model
        for a given atomic configuration.

        Parameters
        ----------
        structure
            input atomic structure.
        """
        return self._cluster_expansion.predict(structure)



import pandas as pd
import numpy as np
import scipy


def analyze_data(data: np.ndarray, max_lag: int = None) -> dict:
    """ Carries out an extensive analysis of the data series.

    Parameters
    ----------
    data
        data series to compute autocorrelation function for
    max_lag
        maximum lag between two data points, used for computing autocorrelation

    Returns
    -------
    dict
        calculated properties of the data including, mean, standard deviation,
        correlation length and a 95% error estimate.
    """
    acf = get_autocorrelation_function(data, max_lag)
    correlation_length = _estimate_correlation_length_from_acf(acf)
    error_estimate = _estimate_error(data, correlation_length, confidence=0.95)
    summary = dict(mean=data.mean(),
                   std=data.std(),
                   correlation_length=correlation_length,
                   error_estimate=error_estimate)
    return summary


def get_autocorrelation_function(data: np.ndarray, max_lag: int = None) -> np.ndarray:
    """ Returns autocorrelation function.

    The autocorrelation function is computed using `pandas.Series.autocorr
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.autocorr.html>`_.

    Parameters
    ----------
    data
        data series to compute autocorrelation function for
    max_lag
        maximum lag between two data points

    Returns
    -------
        calculated autocorrelation function
    """
    if max_lag is None:
        max_lag = len(data) - 1
    if 1 > max_lag >= len(data):
        raise ValueError('max_lag should be between 1 and len(data)-1.')
    series = pd.Series(data)
    acf = [series.autocorr(lag) for lag in range(0, max_lag)]
    return np.array(acf)


def get_correlation_length(data: np.ndarray) -> int:
    """ Returns estimate of the correlation length of data.

    The correlation length is taken as the first point where the
    autocorrelation functions is less than :math:`\\exp(-2)`. If the
    correlation function never drops below :math:`\\exp(-2)` ``np.nan`` is
    returned.

    Parameters
    ----------
    data
        data series for which to the compute autocorrelation function

    Returns
    -------
        correlation length
    """

    acf = get_autocorrelation_function(data)
    correlation_length = _estimate_correlation_length_from_acf(acf)
    return correlation_length


def get_error_estimate(data: np.ndarray, confidence: float = 0.95) -> float:
    """ Returns estimate of standard error :math:`\\mathrm{error}`
    with confidence interval.

    .. math::

       \\mathrm{error} = t_\\mathrm{factor} * \\mathrm{std}(\\mathrm{data}) / \\sqrt{N_s}

    where :math:`t_{factor}` is the factor corresponding to the confidence
    interval and :math:`N_s` is the number of independent measurements
    (with correlation taken into account).

    Parameters
    ----------
    data
        data series for which to estimate the error

    Returns
    -------
        error estimate
    """
    correlation_length = get_correlation_length(data)
    error_estimate = _estimate_error(data, correlation_length, confidence)
    return error_estimate


def _estimate_correlation_length_from_acf(acf: np.ndarray) -> int:
    """ Estimates correlation length from acf. """
    for i, a in enumerate(acf):
        if a < np.exp(-2):
            return i
    return np.nan


def _estimate_error(data: np.ndarray, correlation_length: int,
                    confidence: float) -> float:
    """ Estimates error using correlation length. """
    t_factor = scipy.stats.t.ppf((1 + confidence) / 2, len(data)-1)
    error = t_factor * np.std(data) / np.sqrt(len(data) / correlation_length)
    return error



# -*- coding: utf-8 -*-
"""
Main module of the icet package.
"""

from __future__ import print_function, division

from .core.cluster_space import ClusterSpace, get_singlet_info, view_singlets
from .core.cluster_expansion import ClusterExpansion
from .core.orbit_list import OrbitList
from .core.structure import Structure
from .core.structure_container import StructureContainer
from .fitting import (Optimizer,
                      EnsembleOptimizer,
                      CrossValidationEstimator)

__project__ = 'icet'
__description__ = 'A Pythonic approach to cluster expansions'
__authors__ = ['Mattias Ångqvist',
               'William A. Muñoz',
               'J. Magnus Rahm',
               'Erik Fransson',
               'Céline Durniak',
               'Piotr Rozyczko',
               'Thomas Holm Rod',
               'Paul Erhart']
__copyright__ = '2019'
__license__ = 'Mozilla Public License 2.0 (MPL 2.0)'
__version__ = '1.1'
__maintainer__ = 'The icet developers team'
__email__ = 'icet@materialsmodeling.org'
__status__ = 'Stable'
__url__ = 'http://icet.materialsmodeling.org/'

__all__ = ['ClusterSpace',
           'ClusterExpansion',
           'Structure',
           'OrbitList',
           'StructureContainer',
           'Optimizer',
           'EnsembleOptimizer',
           'CrossValidationEstimator',
           'get_singlet_info',
           'view_singlets']



from typing import List, Tuple

import numpy as np

from ase import Atoms
from ase.build import make_supercell
from ase.geometry import get_distances
from scipy.optimize import linear_sum_assignment
from icet.io.logging import logger


def map_structure_to_reference(relaxed: Atoms,
                               reference: Atoms,
                               inert_species: List[str] = None,
                               tol_positions: float = 1e-4,
                               suppress_warnings: bool = False,
                               assume_no_cell_relaxation: bool = False) \
        -> Tuple[Atoms, dict]:
    """Maps a relaxed structure onto a reference structure.
    The function returns a tuple comprising the ideal supercell most
    closely matching the relaxed structure and a dictionary with
    supplementary information concerning the mapping. The latter
    includes for example the largest deviation of any position in the
    relaxed structure from its reference position (`drmax`), the average
    deviation of the positions in the relaxed structure from the
    reference positions (`dravg`), and the strain tensor for the relaxed
    structure relative to the reference structure (`strain_tensor`).

    The Atoms object that provide further supplemental information via
    custom per-atom arrays including the atomic displacements
    (`Displacement`, `Displacement_Magnitude`) as well as the
    distances to the three closest sites (`Minimum_Distances`).

    Parameters
    ----------
    relaxed
        relaxed input structure
    reference
        reference structure, which can but need not be the primitive structure
    inert_species
        list of chemical symbols (e.g., ``['Au', 'Pd']``) that are never
        substituted for a vacancy; the number of inert sites is used to rescale
        the volume of the relaxed structure to match the reference structure.
    tol_positions
        tolerance factor applied when scanning for overlapping positions in
        Angstrom (forwarded to :func:`ase.build.make_supercell`)
    suppress_warnings
        if True, print no warnings of large strain or relaxation distances
    assume_no_cell_relaxation
        if False volume and cell metric of the relaxed structure are rescaled
        to match the reference structure; this can be unnecessary (and
        counterproductive) for some structures, e.g., with many vacancies

        **Note**: When setting this parameter to False the reference cell metric
        must be obtainable via an *integer* transformation matrix from the
        reference cell metric. In other words the relaxed structure should not
        involve relaxations of the volume or the cell metric.

    Example
    -------
    The following code snippet illustrates the general usage. It first creates
    a primitive FCC cell, which is latter used as reference structure. To
    emulate a relaxed structure obtained from, e.g., a density functional
    theory calculation, the code then creates a 4x4x4 conventional FCC
    supercell, which is populated with two different atom types, has distorted
    cell vectors, and random displacements to the atoms. Finally, the present
    function is used to map the structure back the ideal lattice::

        from icet.tools import map_structure_to_reference
        from ase.build import bulk
        reference = bulk('Au', a=4.09)
        structure = bulk('Au', cubic=True, a=4.09).repeat(4)
        structure.set_chemical_symbols(10 * ['Ag'] + (len(structure) - 10) * ['Au'])
        structure.set_cell(structure.cell * 1.02, scale_atoms=True)
        structure.rattle(0.1)
        mapped_structure, info = map_structure_to_reference(structure, reference)

    """

    # Obtain supercell of reference structure that is compatible
    # with relaxed structure
    reference_supercell = _get_reference_supercell(
        relaxed, reference, inert_species=inert_species,
        tol_positions=tol_positions, assume_no_cell_relaxation=assume_no_cell_relaxation)

    # Compute strain
    epsilon = np.dot(relaxed.cell.T - reference_supercell.cell.T,
                     np.linalg.inv(reference_supercell.cell.T))
    strain_tensor = 0.5 * (epsilon + epsilon.T)
    eigenvalues = np.linalg.eig(strain_tensor)[0]
    volumetric_strain = sum(eigenvalues)

    # Rescale the relaxed atoms object
    relaxed_scaled = relaxed.copy()
    relaxed_scaled.set_cell(reference_supercell.cell, scale_atoms=True)

    # Match positions
    mapped_structure, drmax, dravg = _match_positions(relaxed_scaled, reference_supercell)

    if not suppress_warnings:
        s = 'Consider excluding this structure when training a cluster expansion.'
        if assume_no_cell_relaxation:
            trigger_levels = {'volumetric_strain': 1e-3,
                              'eigenvalue_diff': 1e-3}
        else:
            trigger_levels = {'volumetric_strain': 0.25,
                              'eigenvalue_diff': 0.1}
        if abs(volumetric_strain) > trigger_levels['volumetric_strain']:
            logger.warning('High volumetric strain ({:.2f} %). {}'.format(
                100 * volumetric_strain, s))
        if max(eigenvalues) - min(eigenvalues) > trigger_levels['eigenvalue_diff']:
            logger.warning('High anisotropic strain (the difference between '
                           'largest and smallest eigenvalues of strain tensor is '
                           '{:.5f}). {}'.format(max(eigenvalues) - min(eigenvalues), s))
        if drmax > 1.0:
            logger.warning('Large maximum relaxation distance '
                           '({:.5f} Angstrom). {}'.format(drmax, s))
        if dravg > 0.5:
            logger.warning('Large average relaxation distance '
                           '({:.5f} Angstrom). {}'.format(dravg, s))

    # Populate dictionary with supplementary information
    info = {'drmax': drmax,
            'dravg': dravg,
            'strain_tensor': strain_tensor,
            'volumetric_strain': volumetric_strain,
            'strain_tensor_eigenvalues': eigenvalues}

    return mapped_structure, info


def _get_reference_supercell(relaxed: Atoms,
                             reference: Atoms,
                             inert_species: List[str] = None,
                             tol_positions: float = 1e-5,
                             assume_no_cell_relaxation: bool = False) -> Atoms:
    """
    Returns a supercell of the reference structure that is compatible with the
    cell metric of the relaxed structure.

    Parameters
    ----------
    relaxed
        relaxed input structure
    reference
        reference structure, which can but need not be the primitive structure
    inert_species
        list of chemical symbols (e.g., ``['Au', 'Pd']``) that are never
        substituted for a vacancy; the number of inert sites is used to rescale
        of the relaxed structure to match the reference structure.
    tol_positions
        tolerance factor applied when scanning for overlapping positions in
        Angstrom (forwarded to :func:`ase.build.make_supercell`)
    assume_no_cell_relaxation
        if False volume and cell metric of the relaxed structure are rescaled
        to match the reference structure; this can be unnecessary (and
        counterproductive) for some structures, e.g., with many vacancies

        **Note**: When setting this parameter to False relaxed structure
        must be obtainable via an *integer* transformation matrix from the
        reference cell metric. In other words the relaxed structure should not
        involve relaxations of the volume or the cell metric.

    Raises
    ------
    ValueError
        if the boundary conditions of the reference and the relaxed structure
        do not match
    ValueError
        if the transformation matrix deviates too strongly from the nearest
        integer matrix
    ValueError
        if assume_no_cell_relaxation is True but the relaxed structure is
        not obtainable via an integer transformation from the reference
        cell metric
    """

    if not np.all(reference.pbc == relaxed.pbc):
        msg = 'The boundary conditions of reference and relaxed structures do not match.'
        msg += '\n  reference: ' + str(reference.pbc)
        msg += '\n  relaxed: ' + str(relaxed.pbc)
        raise ValueError(msg)

    # Step 1:
    # rescale cell metric of relaxed cell to match volume per atom of reference cell
    if not assume_no_cell_relaxation:
        if inert_species is None:
            n_ref = len(reference)
            n_rlx = len(relaxed)
        else:
            n_ref = sum([reference.get_chemical_symbols().count(s) for s in inert_species])
            n_rlx = sum([relaxed.get_chemical_symbols().count(s) for s in inert_species])
        vol_scale = reference.get_volume() / n_ref
        vol_scale /= relaxed.get_volume() / n_rlx
        scaled_relaxed_cell = relaxed.cell * vol_scale ** (1 / 3)
    else:
        scaled_relaxed_cell = relaxed.cell

    # Step 2:
    # get transformation matrix
    P = np.dot(scaled_relaxed_cell, np.linalg.inv(reference.cell))

    # reduce the (real) transformation matrix to the nearest integer one
    P = np.around(P)

    # Step 3:
    # generate supercell of reference structure
    reference_supercell = make_supercell(reference, P, tol=tol_positions)

    return reference_supercell


def _match_positions(relaxed: Atoms, reference: Atoms) -> Tuple[Atoms, float, float]:
    """Matches the atoms in the `relaxed` structure to the sites in the
    `reference` structure. The function returns tuple the first element of which
    is a copy of the `reference` structure, in which the chemical species are
    assigned to comply with the `relaxed` structure. The second and third element
    of the tuple represent the maximum and average distance between relaxed and
    reference sites.

    Parameters
    ----------
    relaxed
        structure with relaxed positions
    reference
        structure with idealized positions

    Raises
    ------
    ValueError
        if the cell metrics of the two input structures do not match
    ValueError
        if the periodic boundary conditions of the two input structures do not match
    ValueError
        if the relaxed structure contains more atoms than the reference structure
    """

    if not np.all(reference.pbc == relaxed.pbc):
        msg = 'The boundary conditions of reference and relaxed structures do not match.'
        msg += '\n  reference: ' + str(reference.pbc)
        msg += '\n  relaxed: ' + str(relaxed.pbc)
        raise ValueError(msg)

    if len(relaxed) > len(reference):
        msg = 'The relaxed structure contains more atoms than the reference structure.'
        msg += '\n  reference: ' + str(len(reference))
        msg += '\n  relaxed: ' + str(len(relaxed))
        raise ValueError(msg)

    if not np.all(np.isclose(reference.cell, relaxed.cell)):
        msg = 'The cell metrics of reference and relaxed structures do not match.'
        msg += '\n  reference: ' + str(reference.cell)
        msg += '\n  relaxed: ' + str(relaxed.cell)
        raise ValueError(msg)

    # compute distances between reference and relaxed positions
    _, dists = get_distances(reference.positions, relaxed.positions,
                             cell=reference.cell, pbc=reference.pbc)
    # pad matrix with zeros to obtain square matrix
    n, m = dists.shape
    cost_matrix = np.pad(dists, ((0, 0), (0, abs(n - m))),
                         mode='constant', constant_values=0)
    # find optimal mapping using Kuhn-Munkres (Hungarian) algorithm
    row_ind, col_ind = linear_sum_assignment(cost_matrix)

    # compile new configuration with supplementary information
    mapped = reference.copy()
    displacement_magnitudes = []
    displacements = []
    minimum_distances = []
    for i, j in zip(row_ind, col_ind):
        atom = mapped[i]
        if j >= len(relaxed):
            # vacant site in reference structure
            atom.symbol = 'X'
            displacement_magnitudes.append(None)
            displacements.append(3 * [None])
            minimum_distances.append(3 * [None])
        else:
            atom.symbol = relaxed[j].symbol
            dvecs, drs = get_distances([relaxed[j].position],
                                       [reference[i].position],
                                       cell=reference.cell, pbc=reference.pbc)
            displacement_magnitudes.append(drs[0][0])
            displacements.append(dvecs[0][0])
            # distances to the next three available sites
            minimum_distances.append(sorted(dists[:, j])[:3])

    displacement_magnitudes = np.array(displacement_magnitudes, dtype=np.float64)
    mapped.new_array('Displacement', displacements, float, (3, ))
    mapped.new_array('Displacement_Magnitude', displacement_magnitudes, float)
    mapped.new_array('Minimum_Distances', minimum_distances,
                     float, (min(3, len(mapped)), ))

    drmax = np.nanmax(displacement_magnitudes)
    dravg = np.nanmean(displacement_magnitudes)

    return mapped, drmax, dravg



# -*- coding: utf-8 -*-

from .canonical_annealing import CanonicalAnnealing
from .canonical_ensemble import CanonicalEnsemble
from .hybrid_ensemble import HybridEnsemble
from .semi_grand_canonical_ensemble import SemiGrandCanonicalEnsemble
from .sgc_annealing import SGCAnnealing
from .target_cluster_vector_annealing import TargetClusterVectorAnnealing
from .vcsgc_ensemble import VCSGCEnsemble
from .wang_landau_ensemble import WangLandauEnsemble

__all__ = ['CanonicalEnsemble',
           'CanonicalAnnealing',
           'HybridEnsemble',
           'SemiGrandCanonicalEnsemble',
           'SGCAnnealing',
           'TargetClusterVectorAnnealing',
           'VCSGCEnsemble',
           'WangLandauEnsemble']



from abc import abstractproperty
from typing import Dict, List

import numpy as np

from ase import Atoms
from ase.units import kB
from ase.data import chemical_symbols

from ..calculators.base_calculator import BaseCalculator
from ..data_containers.base_data_container import BaseDataContainer
from .base_ensemble import BaseEnsemble


class ThermodynamicBaseEnsemble(BaseEnsemble):
    """
    Thermodynamic base ensemble class.

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of structure of different species
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 user_tag: str = None,
                 boltzmann_constant: float = kB,
                 random_seed: int = None,
                 data_container: str = None,
                 data_container_class: BaseDataContainer = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None) -> None:

        self._boltzmann_constant = boltzmann_constant

        super().__init__(
            structure=structure,
            calculator=calculator,
            user_tag=user_tag,
            random_seed=random_seed,
            data_container=data_container,
            data_container_class=data_container_class,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval)

    @abstractproperty
    @property
    def temperature(self) -> float:
        pass

    @property
    def boltzmann_constant(self) -> float:
        """ Boltzmann constant :math:`k_B` (see parameters section above) """
        return self._boltzmann_constant

    def _acceptance_condition(self, potential_diff: float) -> bool:
        """
        Evaluates Metropolis acceptance criterion.

        Parameters
        ----------
        potential_diff
            change in the thermodynamic potential associated
            with the trial step
        """
        if potential_diff <= 0:
            return True
        elif self.temperature <= 1e-16:
            return False
        else:
            p = np.exp(-potential_diff / (self.boltzmann_constant * self.temperature))
            return p > self._next_random_number()

    def do_canonical_swap(self, sublattice_index: int, allowed_species: List[int] = None) -> int:
        """ Carries out one Monte Carlo trial step.

        Parameters
        ---------
        sublattice_index
            the sublattice the swap will act on
        allowed_species
            list of atomic numbers for allowed species

        Returns
        -------
        Returns 1 or 0 depending on if trial move was accepted or rejected
        """
        sites, species = self.configuration.get_swapped_state(sublattice_index, allowed_species)
        potential_diff = self._get_property_change(sites, species)

        if self._acceptance_condition(potential_diff):
            self.update_occupations(sites, species)
            return 1
        return 0

    def do_sgc_flip(self, chemical_potentials: Dict[int, float], sublattice_index: int,
                    allowed_species: List[int] = None) -> int:
        """ Carries out one Monte Carlo trial step.

        Parameters
        ---------
        chemical_potentials
            chemical potentials used to calculate the potential
             difference
        sublattice_index
            the sublattice the flip will act on
        allowed_species
            list of atomic numbers for allowed species

        Returns
        -------
        Returns 1 or 0 depending on if trial move was accepted or rejected
        """
        index, species = self.configuration.get_flip_state(sublattice_index, allowed_species)
        potential_diff = self._get_property_change([index], [species])

        # change in chemical potential
        old_species = self.configuration.occupations[index]
        chemical_potential_diff = chemical_potentials[old_species] - chemical_potentials[species]
        potential_diff += chemical_potential_diff

        if self._acceptance_condition(potential_diff):
            self.update_occupations([index], [species])
            return 1
        return 0

    def do_vcsgc_flip(self, phis: Dict[int, float], kappa: float, sublattice_index: int,
                      allowed_species: List[int] = None) -> int:
        """Carries out one Monte Carlo trial step.

        Parameters
        ----------
        phis
            average constraint parameters
        kappa
            parameter that constrains the variance of the concentration
        sublattice_index
            the sublattice the flip will act on
        allowed_species
            list of atomic numbers for allowed species

        Returns
        -------
        Returns 1 or 0 depending on if trial move was accepted or rejected
        """
        index, new_species = self.configuration.get_flip_state(
            sublattice_index, allowed_species)
        old_species = self.configuration.occupations[index]

        # Calculate difference in VCSGC thermodynamic potential.
        # Note that this assumes that only one atom was flipped.
        sl_occupations = self.configuration.get_occupations_on_sublattice(sublattice_index)
        N = len(sl_occupations)
        potential_diff = 1.0  # dN
        for species in phis:
            if species == old_species:
                factor = -1
            elif species == new_species:
                factor = 1
            else:
                continue
            potential_diff += factor * (N * phis[species] + 2 * sl_occupations.count(species))
        potential_diff *= kappa * self.boltzmann_constant * self.temperature / N
        potential_diff += self._get_property_change([index], [new_species])

        if self._acceptance_condition(potential_diff):
            self.update_occupations([index], [new_species])
            return 1
        return 0

    def _get_swap_sublattice_probabilities(self) -> List[float]:
        """ Returns sublattice probabilities suitable for swaps."""
        sublattice_probabilities = []
        for i, sl in enumerate(self.sublattices):
            if self.configuration.is_swap_possible(i):
                sublattice_probabilities.append(len(sl.indices))
            else:
                sublattice_probabilities.append(0)
        norm = sum(sublattice_probabilities)
        if norm == 0:
            raise ValueError('No canonical swaps are possible on any of the active sublattices.')
        sublattice_probabilities = [p / norm for p in sublattice_probabilities]
        return sublattice_probabilities

    def _get_flip_sublattice_probabilities(self) -> List[float]:
        """Returns the default sublattice probability which is based on
        the sizes of a sublattice.
        """
        probability_distribution = []
        for i, sl in enumerate(self.sublattices):
            if len(sl.chemical_symbols) > 1:
                probability_distribution.append(len(sl.indices))
            else:
                probability_distribution.append(0)

        norm = sum(probability_distribution)
        probability_distribution = [p / norm for p in probability_distribution]
        return probability_distribution

    def _get_vcsgc_free_energy_derivatives(self, phis: Dict[int, str], kappa: float,
                                           sublattice_index: int = None) -> Dict:
        """
        Returns a dict with the free energy derivatives.

        Parameters
        ----------
        phis
            average constraint parameters
        kappa
            parameter that constrains the variance of the concentration
        sublattice_index
            sublattice index
        """
        data = {}

        for atnum in phis:
            for i, sublattice in enumerate(self.sublattices):
                if sublattice_index is not None and i != sublattice_index:
                    continue
                if len(sublattice.chemical_symbols) > 0 and atnum in sublattice.atomic_numbers:
                    N = len(sublattice.indices)
                    sl_occupations = self.configuration.get_occupations_on_sublattice(i)
                    concentration = sl_occupations.count(atnum) / N
                    data['free_energy_derivative_{}'.format(chemical_symbols[atnum])] \
                        = kappa * self.boltzmann_constant * self.temperature * \
                        (- 2 * concentration - phis[atnum])

        return data

    def _get_species_counts(self) -> Dict:
        """
        Returns a dict with the species counts.
        """
        data = {}
        structure = self.configuration.structure
        unique, counts = np.unique(structure.numbers, return_counts=True)
        for sl in self.sublattices:
            for symbol in sl.chemical_symbols:
                data['{}_count'.format(symbol)] = 0
        for atnum, count in zip(unique, counts):
            data['{}_count'.format(chemical_symbols[atnum])] = count

        return data



"""Definition of the Wang-Landau algorithm class."""

import random

from collections import OrderedDict
from typing import BinaryIO, Dict, List, TextIO, Union

import numpy as np

from ase import Atoms
from pandas import DataFrame

from .. import WangLandauDataContainer
from ..calculators.base_calculator import BaseCalculator
from .thermodynamic_base_ensemble import BaseEnsemble


class WangLandauEnsemble(BaseEnsemble):
    """Instances of this class allow one to sample a system using the
    Wang-Landau (WL) algorithm, see Phys. Rev. Lett. **86**, 2050
    (2001) [WanLan01a]_. The WL algorithm enables one to acquire the
    density of states (DOS) as a function of energy, from which one
    can readily calculate many thermodynamic observables as a function
    of temperature. To this end, the WL algorithm accumulates both the
    microcanonical entropy :math:`S(E)` and a histogram :math:`H(E)`
    on an energy grid with a predefined spacing (``energy_spacing``).

    The algorithm is initialized as follows.

     #. Generate an initial configuration.
     #. Initialize counters for the microcanonical entropy
        :math:`S(E)` and the histogram :math:`H(E)` to zero.
     #. Set the fill factor :math:`f=1`.


    It then proceeds as follows.

    #. Propose a new configuration (see ``trial_move``).
    #. Accept or reject the new configuration with probability

       .. math::

          P = \\min \\{ 1, \\, \\exp [ S(E_\\mathrm{new}) - S(E_\\mathrm{cur}) ] \\},

       where :math:`E_\\mathrm{cur}` and :math:`E_\\mathrm{new}` are the
       energies of the current and new configurations, respectively.
    #. Update the microcanonical entropy :math:`S(E)\\leftarrow S(E) + f`
       and histogram :math:`H(E) \\leftarrow H(E) + 1` where
       :math:`E` is the energy of the system at the end of the move.
    #. Check the flatness of the histogram :math:`H(E)`. If
       :math:`H(E) > \\chi \\langle H(E)\\rangle\\,\\forall E` reset the histogram
       :math:`H(E) = 0` and reduce the fill factor :math:`f \\leftarrow f / 2`.
       The parameter :math:`\\chi` is set via ``flatness_limit``.
    #. If :math:`f` is smaller than ``fill_factor_limit`` terminate
       the loop, otherwise return to 1.

    The microcanonical entropy :math:`S(E)` and the histogram along
    with related information are written to the data container every
    time :math:`f` is updated. Using the density :math:`\\rho(E) = \\exp S(E)`
    one can then readily compute various thermodynamic quantities,
    including, e.g., the average energy:

    .. math::

       \\left<E\\right> = \\frac{\\sum_E E \\rho(E) \\exp(-E / k_B T)}{
       \\sum_E \\rho(E) \\exp(-E / k_B T)}

    Parameters
    ----------
    structure : :class:`Atoms <ase.Atoms>`
        atomic configuration to be used in the Wang-Landau simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator <mchammer.calculators.ClusterExpansionCalculator>`
        calculator to be used for calculating potential changes
    trial_move : str
        One can choose between two different trial moves for
        generating new configurations. In a 'swap' move two sites are
        selected and their occupations are swapped; in a 'flip' move
        one site is selected and its occupation is flipped to a
        different species. While 'swap' moves conserve the
        concentrations of the species in the system, 'flip' moves
        allow one in principle to sample the full composition space.
    energy_spacing : float
        defines the bin size of the energy grid on which the microcanonical
        entropy :math:`S(E)`, and thus the density :math:`\\exp S(E)`, is
        evaluated; the spacing should be small enough to capture the features
        of the density of states; too small values will, however, render the
        convergence very tedious if not impossible
    energy_limit_left : float
        defines the lower limit of the energy range within which the
        microcanonical entropy :math:`S(E)` will be sampled. By default
        (`None`) no limit is imposed. Setting limits can be useful if only a
        part of the density of states is required.
    energy_limit_right : float
        defines the upper limit of the energy range within which the
        microcanonical entropy :math:`S(E)` will be sampled. By default
        (`None`) no limit is imposed. Setting limits can be useful if only a
        part of the density of states is required.
    fill_factor_limit : float
        If the fill_factor :math:`f` falls below this value, the
        WL sampling loop is terminated.
    flatness_check_interval : int
        For computational efficiency the flatness condition is only
        evaluated every ``flatness_check_interval``-th trial step. By
        default (``None``) ``flatness_check_interval`` is set to 1000
        times the number of sites in ``structure``, i.e. 1000 Monte
        Carlo sweeps.
    flatness_limit : float
        The histogram :math:`H(E)` is deemed sufficiently flat if
        :math:`H(E) > \\chi \\left<H(E)\\right>\\,\\forall
        E`. ``flatness_limit`` sets the parameter :math:`\\chi`.
    user_tag : str
        human-readable tag for ensemble [default: None]
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    sublattice_probabilities : List[float]
        probability for picking a sublattice when doing a random swap.
        The list must contain as many elements as there are sublattices
        and it needs to sum up to 1.

    Examples
    --------
    The following snippet illustrates how to carry out a Wang-Landau
    simulation. For the purpose of demonstration, the parameters of
    the cluster expansion are set to obtain a two-dimensional square
    Ising model, one of the systems studied in the original work by
    Wang and Landau::

        from ase import Atoms
        from icet import ClusterExpansion, ClusterSpace
        from mchammer.calculators import ClusterExpansionCalculator
        from mchammer.ensembles import WangLandauEnsemble

        # prepare cluster expansion
        prim = Atoms('Au', positions=[[0, 0, 0]], cell=[1, 1, 10], pbc=True)
        cs = ClusterSpace(prim, cutoffs=[1.1], chemical_symbols=['Ag', 'Au'])
        ce = ClusterExpansion(cs, [0, 0, 2])

        # prepare initial configuration
        structure = prim.repeat((4, 4, 1))
        for k in range(8):
            structure[k].symbol = 'Ag'

        # set up and run Wang-Landau simulation
        calculator = ClusterExpansionCalculator(structure, ce)
        mc = WangLandauEnsemble(structure=structure,
                                calculator=calculator,
                                energy_spacing=1,
                                data_container='ising_2d_run.dc')
        mc.run(number_of_trial_steps=len(structure)*10000)  # carry out 10,000 MC sweeps

    """

    def __init__(self,
                 structure: Atoms,
                 calculator: BaseCalculator,
                 energy_spacing: float,
                 energy_limit_left: float = None,
                 energy_limit_right: float = None,
                 trial_move: str = 'swap',
                 fill_factor_limit: float = 1e-6,
                 flatness_check_interval: int = None,
                 flatness_limit: float = 0.8,
                 user_tag: str = None,
                 data_container: str = None,
                 random_seed: int = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None,
                 sublattice_probabilities: List[float] = None) -> None:

        # set trial move
        if trial_move == 'swap':
            self.do_move = self._do_swap
            self._get_sublattice_probabilities = self._get_swap_sublattice_probabilities
        elif trial_move == 'flip':
            self.do_move = self._do_flip
            self._get_sublattice_probabilities = self._get_flip_sublattice_probabilities
        else:
            raise ValueError('Invalid value for trial_move: {}.'
                             ' Must be either "swap" or "flip".'.format(trial_move))

        # set default values that are system dependent
        if flatness_check_interval is None:
            flatness_check_interval = len(structure) * 1e3

        # parameters pertaining to construction of entropy and histogram
        self._energy_spacing = energy_spacing
        self._fill_factor_limit = fill_factor_limit
        self._flatness_check_interval = flatness_check_interval
        self._flatness_limit = flatness_limit

        # energy window
        self._bin_left = self._get_bin_index(energy_limit_left)
        self._bin_right = self._get_bin_index(energy_limit_right)
        if self._bin_left is not None and \
                self._bin_right is not None and self._bin_left >= self._bin_right:
            raise ValueError('Invalid energy window: left boundary ({}, {}) must be'
                             ' smaller than right boundary ({}, {})'
                             .format(energy_limit_left, self._bin_left,
                                     energy_limit_right, self._bin_right))
        self._reached_energy_window = self._bin_left is None and self._bin_right is None

        # ensemble parameters
        self._ensemble_parameters = {}
        self._ensemble_parameters['energy_spacing'] = energy_spacing
        self._ensemble_parameters['trial_move'] = trial_move
        self._ensemble_parameters['energy_limit_left'] = energy_limit_left
        self._ensemble_parameters['energy_limit_right'] = energy_limit_right
        # The following parameters are _intentionally excluded_ from
        # the ensemble_parameters dict as it would prevent users from
        # changing their values between restarts. The latter is advantageous
        # as these runs can require restarts and possibly parameter adjustments
        # to achieve convergence.
        #  * fill_factor_limit
        #  * flatness_check_interval
        #  * flatness_limit

        # add species count to ensemble parameters
        symbols = set([symbol for sub in calculator.sublattices
                       for symbol in sub.chemical_symbols])
        for symbol in symbols:
            key = 'n_atoms_{}'.format(symbol)
            count = structure.get_chemical_symbols().count(symbol)
            self._ensemble_parameters[key] = count

        # the constructor of the parent classes must be called *after*
        # the ensemble_parameters dict has been populated
        super().__init__(
            structure=structure,
            calculator=calculator,
            user_tag=user_tag,
            random_seed=random_seed,
            data_container=data_container,
            data_container_class=WangLandauDataContainer,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval)

        # handle probabilities for swaps on different sublattices
        if sublattice_probabilities is None:
            self._sublattice_probabilities = self._get_sublattice_probabilities()
        else:
            self._sublattice_probabilities = sublattice_probabilities

        # initialize Wang-Landau algorithm; in the case of a restart
        # these quantities are read from the data container file; the
        # if-conditions prevent these values from being overwritten
        self._converged = None
        self._potential = self.calculator.calculate_total(
            occupations=self.configuration.occupations)
        if not hasattr(self, '_fill_factor'):
            self._fill_factor = 1
        if not hasattr(self, '_fill_factor_history'):
            if self._reached_energy_window:
                self._fill_factor_history = {self.step: self._fill_factor}
            else:
                self._fill_factor_history = {}
        if not hasattr(self, '_histogram'):
            self._histogram = {}
        if not hasattr(self, '_entropy'):
            self._entropy = {}

    @property
    def fill_factor(self) -> float:
        """ current value of fill factor """
        return self._fill_factor

    @property
    def fill_factor_history(self) -> Dict[int, float]:
        """evolution of the fill factor in the Wang-Landau algorithm (key=MC
        trial step, value=fill factor)
        """
        return self._fill_factor_history

    @property
    def converged(self) -> bool:
        """ True if convergence has been achieved """
        return self._converged

    @property
    def flatness_limit(self) -> float:
        """The histogram :math:`H(E)` is deemed sufficiently flat if
        :math:`H(E) > \\chi \\left<H(E)\\right>\\,\\forall
        E` where ``flatness_limit`` sets the parameter :math:`\\chi`.
        """
        return self._flatness_limit

    @flatness_limit.setter
    def flatness_limit(self, new_value) -> None:
        self._flatness_limit = new_value
        self._converged = None

    @property
    def fill_factor_limit(self) -> float:
        """ If the fill_factor :math:`f` falls below this value, the
        Wang-Landau sampling loop is terminated. """
        return self._fill_factor_limit

    @fill_factor_limit.setter
    def fill_factor_limit(self, new_value) -> None:
        self._fill_factor_limit = new_value
        self._converged = None

    @property
    def flatness_check_interval(self) -> int:
        """ number of MC trial steps between checking the flatness
        condition """
        return self._flatness_check_interval

    @flatness_check_interval.setter
    def flatness_check_interval(self, new_value: int) -> None:
        self._flatness_check_interval = new_value

    def get_entropy(self) -> DataFrame:
        """ entropy accumulated during Wang-Landau simulation """
        df = DataFrame(data={'energy': self._energy_spacing * np.array(list(self._entropy.keys())),
                             'entropy': np.array(list(self._entropy.values()))},
                       index=list(self._entropy.keys()))
        df['entropy'] -= np.min(df['entropy'])
        return df

    def get_histogram(self) -> DataFrame:
        """ histogram accumulated since last update of fill factor """
        df = DataFrame(data={'energy': self._energy_spacing *
                             np.array(list(self._histogram.keys())),
                             'histogram': np.array(list(self._histogram.values()))},
                       index=list(self._histogram.keys()))
        df['histogram'] -= np.min(df['histogram'])
        return df

    def _terminate_sampling(self) -> bool:
        """Returns True if the Wang-Landau algorithm has converged. This is
        used in the run method implemented of BaseEnsemble to
        evaluate whether the sampling loop should be terminated.
        """
        return self._converged is True  # N.B.: self._converged can be None

    def _restart_ensemble(self):
        """Restarts ensemble using the last state saved in the data container
        file. Note that this method does _not_ use the last_state property of
        the data container but rather uses the last data written the data frame.
        """
        super()._restart_ensemble()
        self._fill_factor = self.data_container.last_state['fill_factor']
        self._fill_factor_history = self.data_container.last_state['fill_factor_history']
        self._histogram = self.data_container.last_state['histogram']
        self._entropy = self.data_container.last_state['entropy']

    def write_data_container(self, outfile: Union[str, BinaryIO, TextIO]):
        """Updates last state of the Wang-Landau simulation and
        writes DataContainer to file.

        Parameters
        ----------
        outfile
            file to which to write
        """
        self._data_container._update_last_state(
            last_step=self.step,
            occupations=self.configuration.occupations.tolist(),
            accepted_trials=self._accepted_trials,
            random_state=random.getstate(),
            fill_factor=self._fill_factor,
            fill_factor_history=self._fill_factor_history,
            histogram=OrderedDict(sorted(self._histogram.items())),
            entropy=OrderedDict(sorted(self._entropy.items())))
        self.data_container.write(outfile)

    def _acceptance_condition(self, potential_diff: float) -> bool:
        """Evaluates Metropolis acceptance criterion.

        Parameters
        ----------
        potential_diff
            change in the thermodynamic potential associated
            with the trial step
        """

        # acceptance/rejection step
        bin_cur = self._get_bin_index(self._potential)
        bin_new = self._get_bin_index(self._potential + potential_diff)
        if self._allow_move(bin_cur, bin_new):
            S_cur = self._entropy.get(bin_cur, 0)
            S_new = self._entropy.get(bin_new, 0)
            delta = np.exp(S_cur - S_new)
            if delta >= 1 or delta >= self._next_random_number():
                accept = True
                self._potential += potential_diff
                bin_cur = bin_new
            else:
                accept = False
        else:
            accept = False

        if not self._reached_energy_window:
            # check whether the target energy window has been reached
            self._reached_energy_window = self._inside_energy_window(bin_cur)
            # if the target window has been reached remove unused bins
            # from histogram and entropy counters
            if self._reached_energy_window:
                self._fill_factor_history[self.step] = self._fill_factor
                # flush data from data container except for initial step
                self._data_container._data_list = [self._data_container._data_list[0]]
                self._entropy = {k: self._entropy[k]
                                 for k in self._entropy if self._inside_energy_window(k)}
                self._histogram = {k: self._histogram[k]
                                   for k in self._histogram if self._inside_energy_window(k)}

        # update histograms and entropy counters
        self._update_entropy(bin_cur)

        return accept

    def _update_entropy(self, bin_cur: int) -> None:
        """Updates counters for histogram and entropy, checks histogram
        flatness, and updates fill factor if indicated.
        """

        # update histogram and entropy
        self._entropy[bin_cur] = self._entropy.get(bin_cur, 0) + self._fill_factor
        self._histogram[bin_cur] = self._histogram.get(bin_cur, 0) + 1

        # check flatness of histogram
        if self.step % self._flatness_check_interval == 0 and \
                self.step > 0 and self._reached_energy_window:

            # shift entropy counter in order to avoid overflow
            entropy_ref = np.min(list(self._entropy.values()))
            for k in self._entropy:
                self._entropy[k] -= entropy_ref

            histogram = np.array(list(self._histogram.values()))
            limit = self._flatness_limit * np.average(histogram)
            if np.all(histogram >= limit):

                # check whether the Wang-Landau algorithm has converged
                self._converged = (self._fill_factor <= self._fill_factor_limit)
                if not self._converged:
                    # update fill factor and reset histogram
                    self._fill_factor /= 2
                    self._fill_factor_history[self.step] = self._fill_factor
                    self._histogram = dict.fromkeys(self._histogram, 0)

    def _get_bin_index(self, energy: float) -> int:
        """ Returns bin index for histogram and entropy dictionaries. """
        if energy is None or np.isnan(energy):
            return None
        return int(np.around(energy / self._energy_spacing))

    def _allow_move(self, bin_cur: int, bin_new: int) -> bool:
        """Returns True if the current move is to be included in the
        accumulation of histogram and entropy. This logic has been
        moved into a separate function in order to enhance
        readability.
        """
        if self._bin_left is None and self._bin_right is None:
            # no limits on energy window
            return True
        if self._bin_left is not None:
            if bin_cur < self._bin_left:
                # not yet in window (left limit)
                return True
            if bin_new < self._bin_left:
                # imposing left limit
                return False
        if self._bin_right is not None:
            if bin_cur > self._bin_right:
                # not yet in window (right limit)
                return True
            if bin_new > self._bin_right:
                # imposing right limit
                return False
        return True

    def _inside_energy_window(self, bin_k: int) -> bool:
        """Returns True if bin_k is inside the energy window specified for
        this simulation.
        """
        if self._bin_left is not None and bin_k < self._bin_left:
            return False
        if self._bin_right is not None and bin_k > self._bin_right:
            return False
        return True

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """
        sublattice_index = self.get_random_sublattice_index(self._sublattice_probabilities)
        return self.do_move(sublattice_index=sublattice_index)

    def _do_swap(self, sublattice_index: int, allowed_species: List[int] = None) -> int:
        """Carries out a Monte Carlo trial that involves swapping the species
        on two sites. This method has been copied from
        ThermodynamicBaseEnsemble.

        Parameters
        ---------
        sublattice_index
            the sublattice the swap will act on
        allowed_species
            list of atomic numbers for allowed species

        Returns
        -------
        Returns 1 or 0 depending on if trial move was accepted or rejected
        """
        sites, species = self.configuration.get_swapped_state(sublattice_index, allowed_species)
        potential_diff = self._get_property_change(sites, species)
        if self._acceptance_condition(potential_diff):
            self.update_occupations(sites, species)
            return 1
        return 0

    def _do_flip(self, sublattice_index: int, allowed_species: List[int] = None) -> int:
        """Carries out one Monte Carlo trial step that involves flipping the
        species on one site. This method has been adapted from
        ThermodynamicBaseEnsemble.

        Parameters
        ---------
        sublattice_index
            the sublattice the flip will act on
        allowed_species
            list of atomic numbers for allowed species

        Returns
        -------
        Returns 1 or 0 depending on if trial move was accepted or rejected

        """
        index, species = self.configuration.get_flip_state(sublattice_index, allowed_species)
        potential_diff = self._get_property_change([index], [species])
        if self._acceptance_condition(potential_diff):
            self.update_occupations([index], [species])
            return 1
        return 0

    def _get_swap_sublattice_probabilities(self) -> List[float]:
        """Returns sublattice probabilities suitable for swaps. This method
        has been copied without modification from ThermodynamicBaseEnsemble.
        """
        sublattice_probabilities = []
        for i, sl in enumerate(self.sublattices):
            if self.configuration.is_swap_possible(i):
                sublattice_probabilities.append(len(sl.indices))
            else:
                sublattice_probabilities.append(0)
        norm = sum(sublattice_probabilities)
        if norm == 0:
            raise ValueError('No swaps are possible on any of the active sublattices.')
        sublattice_probabilities = [p / norm for p in sublattice_probabilities]
        return sublattice_probabilities

    def _get_flip_sublattice_probabilities(self) -> List[float]:
        """Returns the default sublattice probability which is based on the
        sizes of a sublattice. This method has been copied without
        modification from ThermodynamicBaseEnsemble.
        """
        sublattice_probabilities = []
        for _, sl in enumerate(self.sublattices):
            if len(sl.chemical_symbols) > 1:
                sublattice_probabilities.append(len(sl.indices))
            else:
                sublattice_probabilities.append(0)
        norm = sum(sublattice_probabilities)
        sublattice_probabilities = [p / norm for p in sublattice_probabilities]
        return sublattice_probabilities
